import os
import unittest

import numpy as np

from gnes.encoder.base import PipelineEncoder


class TestDumpAndLoad(unittest.TestCase):
    def setUp(self):
        dirname = os.path.dirname(__file__)
        self.lopq_yaml_np = os.path.join(dirname, 'yaml', 'lopq-encoder-2-np.yml')
        self.dump_path = os.path.join(dirname, 'encoder.bin')
        self.test_vecs = np.random.random([1000, 100]).astype('float32')

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)

    def _test(self, yaml_fp):
        lopq = PipelineEncoder.load_yaml(yaml_fp)
        lopq.train(self.test_vecs)
        out = lopq.encode(self.test_vecs)
        lopq.dump(self.dump_path)
        self.assertTrue(os.path.exists(self.dump_path))
        lopq2 = PipelineEncoder.load(self.dump_path)
        out2 = lopq2.encode(self.test_vecs)
        self.assertEqual(out, out2)

    def test_dumpload_np(self):
        self._test(self.lopq_yaml_np)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from collections import defaultdict
from typing import Dict, List

from .base import BaseService as BS, MessageHandler, BlockMessage
from ..proto import gnes_pb2
from ..router.base import BaseReduceRouter


class RouterService(BS):
    handler = MessageHandler(BS.handler)

    def post_init(self):
        from ..router.base import BaseRouter
        self._model = self.load_model(BaseRouter)
        self._pending = defaultdict(list)  # type: Dict[str, List]

    def _is_msg_complete(self, msg: 'gnes_pb2.Message', num_req: int) -> bool:
        return (num_req == msg.envelope.num_part[-1]) or \
               (num_req == self.args.num_part)

    @handler.register(NotImplementedError)
    def _handler_default(self, msg: 'gnes_pb2.Message'):
        if isinstance(self._model, BaseReduceRouter):
            req_id = msg.envelope.request_id
            self._pending[req_id].append(msg)
            num_req = len(self._pending[req_id])

            if self._is_msg_complete(msg, num_req):
                prev_msgs = self._pending.pop(req_id)
                return self._model.apply(msg, prev_msgs)
            else:
                raise BlockMessage
        else:
            return self._model.apply(msg)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections import defaultdict
from typing import List

from ..base import BaseReduceRouter
from ...proto import gnes_pb2


class ChunkToDocumentRouter(BaseReduceRouter):

    def apply(self, msg: 'gnes_pb2.Message', accum_msgs: List['gnes_pb2.Message'], *args, **kwargs):
        all_chunks = [r for m in accum_msgs for r in m.response.search.topk_results]
        doc_score = defaultdict(float)
        doc_score_explained = defaultdict(str)

        # count score by iterating over chunks
        for c in all_chunks:
            doc_score[c.chunk.doc_id] += c.score
            doc_score_explained[c.chunk.doc_id] += '%s\n' % c.score_explained

        # now convert chunk results to doc results
        msg.response.search.level = gnes_pb2.Response.QueryResponse.DOCUMENT_NOT_FILLED
        msg.response.search.ClearField('topk_results')

        # sort and add docs
        for k, v in sorted(doc_score.items(), key=lambda kv: -kv[1]):
            r = msg.response.search.topk_results.add()
            r.doc.doc_id = k
            r.score = v
            r.score_explained = doc_score_explained[k]

        super().apply(msg, accum_msgs)



import os
import time
import unittest.mock

import grpc

from gnes.cli.parser import set_frontend_parser, set_router_service_parser
from gnes.helper import TimeContext
from gnes.proto import RequestGenerator, gnes_pb2_grpc
from gnes.service.base import SocketType, MessageHandler, BaseService as BS
from gnes.service.frontend import FrontendService
from gnes.service.router import RouterService


class Router1(RouterService):
    handler = MessageHandler(BS.handler)

    @handler.register(NotImplementedError)
    def _handler_default(self, msg: 'gnes_pb2.Message'):
        self.logger.info('im doing fancy jobs...')
        time.sleep(2)
        super()._handler_default(msg)


class Router2(RouterService):
    handler = MessageHandler(BS.handler)

    @handler.register(NotImplementedError)
    def _handler_default(self, msg: 'gnes_pb2.Message'):
        self.logger.info('im doing stupid jobs...')
        time.sleep(6)
        super()._handler_default(msg)


class TestStreamgRPC(unittest.TestCase):

    def setUp(self):
        self.all_bytes = [b'abc', b'def', b'cde'] * 10
        self.all_bytes2 = [b'abc', b'def', b'cde']
        os.unsetenv('http_proxy')
        os.unsetenv('https_proxy')

    def test_grpc_frontend(self):
        args = set_frontend_parser().parse_args([
            '--grpc_host', '127.0.0.1',
        ])

        p_args = set_router_service_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.PULL_CONNECT),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--yaml_path', 'BaseRouter'
        ])

        with RouterService(p_args), FrontendService(args), grpc.insecure_channel(
                '%s:%s' % (args.grpc_host, args.grpc_port),
                options=[('grpc.max_send_message_length', 70 * 1024 * 1024),
                         ('grpc.max_receive_message_length', 70 * 1024 * 1024)]) as channel:
            stub = gnes_pb2_grpc.GnesRPCStub(channel)
            with TimeContext('sync call'):  # about 5s
                resp = list(stub.StreamCall(RequestGenerator.train(self.all_bytes, 1)))[-1]

            self.assertEqual(resp.request_id, str(len(self.all_bytes)))  # idx start with 0, but +1 for final FLUSH

    def test_async_block(self):
        args = set_frontend_parser().parse_args([
            '--grpc_host', '127.0.0.1',
        ])

        p1_args = set_router_service_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', '8899',
            '--socket_in', str(SocketType.PULL_CONNECT),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--yaml_path', 'BaseRouter'
        ])

        p2_args = set_router_service_parser().parse_args([
            '--port_in', str(p1_args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.PULL_BIND),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--yaml_path', 'BaseRouter'
        ])

        with FrontendService(args), Router1(p1_args), Router2(p2_args), grpc.insecure_channel(
                '%s:%s' % (args.grpc_host, args.grpc_port),
                options=[('grpc.max_send_message_length', 70 * 1024 * 1024),
                         ('grpc.max_receive_message_length', 70 * 1024 * 1024)]) as channel:
            stub = gnes_pb2_grpc.GnesRPCStub(channel)
            id = 0
            with TimeContext('non-blocking call'):  # about 26s = 32s (total) - 3*2s (overlap)
                resp = stub.StreamCall(RequestGenerator.train(self.all_bytes2, 1))
                for r in resp:
                    self.assertEqual(r.request_id, str(id))
                    id += 1

            id = 0
            with TimeContext('blocking call'):  # should be 32 s
                for r in RequestGenerator.train(self.all_bytes2, 1):
                    resp = stub.Call(r)
                    self.assertEqual(resp.request_id, str(id))
                    id += 1
            # self.assertEqual(resp.result().request_id, str(len(self.all_bytes)))

            # self.assertEqual(resp.request_id, str(len(self.all_bytes2)))  # idx start with 0, but +1 for final FLUSH



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




from typing import Dict, Any

import numpy as np

from .pq import PQEncoder
from ...helper import batching, train_required, get_first_available_gpu


class TFPQEncoder(PQEncoder):
    batch_size = 8192

    @classmethod
    def pre_init(cls):
        import os
        os.environ['CUDA_VISIBLE_DEVICES'] = str(get_first_available_gpu())

    def post_init(self):
        import tensorflow as tf
        self._graph = self._get_graph()
        self._sess = tf.Session()
        self._sess.run(tf.global_variables_initializer())

    def _get_graph(self) -> Dict[str, Any]:
        import tensorflow as tf
        ph_x = tf.placeholder(tf.float32, [None, self.num_bytes, None])
        ph_centroids = tf.placeholder(tf.float32, [1, self.num_bytes, self.num_clusters, None])
        centroids_squeezed = tf.squeeze(ph_centroids, 0)
        # [self.num_bytes, None, self.m]
        x = tf.transpose(ph_x, [1, 0, 2])
        ty = tf.reduce_sum(tf.square(centroids_squeezed), axis=2, keepdims=True)
        ty = tf.transpose(ty, [0, 2, 1])
        tx = tf.reduce_sum(tf.square(x), axis=2, keepdims=True)
        diff = tf.matmul(x, tf.transpose(centroids_squeezed, [0, 2, 1]))
        diff = tx + ty - 2 * diff
        # start from 1
        p = tf.argmax(-diff, axis=2) + 1
        p = tf.transpose(p, [1, 0])
        return {
            'out': p,
            'ph_x': ph_x,
            'ph_centroids': ph_centroids
        }

    @train_required
    @batching
    def encode(self, vecs: np.ndarray, *args, **kwargs) -> np.ndarray:
        vecs = np.reshape(vecs, [vecs.shape[0], self.num_bytes, -1])
        tmp = self._sess.run(self._graph['out'],
                             feed_dict={self._graph['ph_x']: vecs,
                                        self._graph['ph_centroids']: self.centroids})
        return tmp.astype(np.uint8)

    def close(self):
        self._sess.close()



import unittest

from gnes.encoder.base import BaseEncoder as BE, PipelineEncoder
from gnes.helper import train_required


class _DummyTrainEncoder(BE):
    @train_required
    def encode(self, *args, **kwargs):
        pass

    def train(self, *args, **kwargs):
        pass


class _BertEncoder(BE):
    def encode(self, data, *args, **kwargs):
        print('bert-encode!')
        return data * 7

    def train(self, data, *args, **kwargs):
        print('bert-train!')


class _PQEncoder(BE):
    def encode(self, data, *args, **kwargs):
        print('pq-encode!')
        return data * 2

    def train(self, data, *args, **kwargs):
        print('pq-train!')


class _PCAEncoder(BE):
    @train_required
    def encode(self, data, *args, **kwargs):
        print('pca-encode!')
        return data + 3

    def train(self, data, *args, **kwargs):
        print('pca-train!')


class _LOPQEncoder(PipelineEncoder):
    def __init__(self):
        super().__init__()
        self.component = lambda: [_PCAEncoder(), _PQEncoder()]


class _BertBinaryEncoder(PipelineEncoder):
    def __init__(self):
        super().__init__()
        self.component = lambda: [_BertEncoder(), _LOPQEncoder()]


class TestDocument(unittest.TestCase):
    def test_no_pretrain(self):
        a = _DummyTrainEncoder()
        self.assertRaises(RuntimeError, a.encode)

    def test_with_pretrain(self):
        a = _DummyTrainEncoder()
        a.train()
        a.encode()

    def test_hierachy_encoder(self):
        le = _LOPQEncoder()
        self.assertRaises(RuntimeError, le.encode, 1)
        le.train(data=1)
        self.assertEqual(le.encode(data=1), 8)
        self.assertEqual(le.encode(data=2), 10)

    def test_hierachy_encoder2(self):
        print('___')
        le2 = _BertBinaryEncoder()
        le2.train(data=1)
        le2.encode(data=1)



# cython: language_level=3, wraparound=False, boundscheck=False, cdivision=True
# noinspection PyUnresolvedReferences
# created 2019-03-27

#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# pylint: disable=low-comment-ratio


from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
from cpython cimport array
from libc.stdlib cimport qsort
from libc.stdio cimport fopen, fclose, FILE, fwrite, fread


cdef extern from "limits.h":
    cdef int USHRT_MAX
    cdef unsigned int UINT_MAX

short_max = USHRT_MAX
int_max = UINT_MAX
DEF Increase = 10000

ctypedef unsigned int uint
ctypedef unsigned short ushort
ctypedef unsigned char uchar

cdef uint bytes_to_uint(uchar*all_ids):
    cdef uint label
    cdef ushort _
    label = 0
    for _ in range(4):
        label += (256**_) * all_ids[_]
    return label

cdef ushort bytes_to_ushort(uchar*all_ids):
    cdef ushort offset
    offset = all_ids[0] + all_ids[1]*256
    return offset


cdef ushort vec_distance(uchar*va, uchar*vb, ushort bytes_per_vector):
    cdef ushort i, dist
    dist = 0
    cdef uchar tmp_vec
    for i in range(bytes_per_vector):
        a = va[i]
        b = vb[i]
        tmp_vec = (a ^ b)

        while tmp_vec > 0:
            dist += tmp_vec & 1
            tmp_vec >>= 1
    return dist

cdef struct res_node:
    uint doc_id
    ushort off_set
    ushort weight
    uint dist


cdef void max_heapify(res_node* ret, int loc, int count):
    cdef int left, right, largest
    cdef uint temp
    cdef ushort temp2, temp3
    left = 2*(loc) + 1
    right = left + 1
    largest = loc

    if left < count:
        if ret[left].dist > ret[largest].dist:
            largest = left

    if right < count:
        if ret[right].dist > ret[largest].dist:
            largest = right

    if(largest != loc):
        temp = ret[loc].doc_id
        ret[loc].doc_id = ret[largest].doc_id
        ret[largest].doc_id = temp

        temp = ret[loc].dist
        ret[loc].dist = ret[largest].dist
        ret[largest].dist = temp

        temp2 = ret[loc].off_set
        ret[loc].off_set = ret[largest].off_set
        ret[largest].off_set = temp2

        temp3 = ret[loc].weight
        ret[loc].weight = ret[largest].weight
        ret[largest].weight = temp3

        max_heapify(ret, largest, count)


cdef void heap_push(res_node* ret, uint value, uint doc_id, ushort off_set, ushort weight, int count):
    cdef int index, parent
    index = count
    if index == 0:
        ret[0].dist = value
        ret[0].doc_id = doc_id
        ret[0].off_set = off_set
        ret[0].weight = weight
    else:
        parent = (index - 1)/2
        while (ret[parent].dist < value):
            if index <= 0:
                break
            ret[index].dist = ret[parent].dist
            ret[index].doc_id = ret[parent].doc_id
            ret[index].off_set = ret[parent].off_set
            ret[index].weight = ret[parent].weight
            index = parent
            parent = (index - 1)/2
        ret[index].dist = value
        ret[index].doc_id = doc_id
        ret[index].off_set = off_set
        ret[index].weight = weight


cdef class IndexCore:
    cdef uchar* vecs
    cdef uint n_clusters
    cdef uint n_bytes
    cdef uint n_idx
    cdef uint i
    cdef uchar initialized
    cdef uchar*** core
    cdef uint** data_num

    def __cinit__(self, n_clusters, n_bytes, n_idx):
        self.n_clusters = n_clusters
        self.n_bytes = n_bytes
        self.n_idx = n_idx
        self.core = <uchar***> PyMem_Malloc(sizeof(uchar**)*n_idx)
        self.data_num = <uint**> PyMem_Malloc(sizeof(uint*)*n_idx)
        for i in range(self.n_idx):
            self.core[i] = <uchar**> PyMem_Malloc(sizeof(uchar*)*self.n_clusters)
        self.initialized = 0

    cdef void _initialize(self):
        cdef uint i, j
        for i in range(self.n_idx):
            for j in range(self.n_clusters):
                self.core[i][j] = <uchar*> PyMem_Malloc(sizeof(uchar)*(self.n_bytes+8)*Increase)
            # initialized the default value of num of data in each array
            self.data_num[i] = <uint*> PyMem_Malloc(sizeof(uint)*self.n_clusters)
            for j in range(self.n_clusters):
                self.data_num[i][j] = 0

    cpdef void index_trie(self, uchar*data, uchar*clusters, uchar*doc_ids, uchar*off_sets, uchar* weights, uint n):
        cdef uint i, clus, num, cur_num
        cdef ushort j, k
        cdef uchar*vec
        if not self.initialized:
            self._initialize()
            self.initialized = 1
        for i in range(n):
            vec = data + 4 * self.n_idx
            for j in range(self.n_idx):
                clus = bytes_to_uint(clusters)
                num = self.data_num[j][clus]
                cur_num = num * (self.n_bytes + 8)
                if (num + 1) % Increase == 0:
                    self.core[j][clus] = <uchar*> PyMem_Realloc(
                        self.core[j][clus],
                        sizeof(uchar)*(self.n_bytes+8)*(num+1+Increase))
                # record doc id
                for k in range(4):
                    self.core[j][clus][cur_num+k] = doc_ids[k]
                cur_num += 4
                # record offset
                for k in range(2):
                    self.core[j][clus][cur_num+k] = off_sets[k]
                cur_num += 2
                for k in range(2):
                    self.core[j][clus][cur_num+k] = weights[k]
                cur_num += 2
                for k in range(self.n_bytes):
                    self.core[j][clus][cur_num+k] = data[k]
                clusters += 4
                self.data_num[j][clus] += 1

            data += self.n_bytes
            doc_ids += 4
            off_sets += 2
            weights += 2

    cpdef query(self, uchar*data, uchar*clusters, uint n, uint top_k):
        cdef array.array res_dist = array.array('L')
        cdef array.array res_docs = array.array('L')
        cdef array.array res_offset = array.array('L')
        cdef array.array res_weight = array.array('L')
        cdef array.array res_idx = array.array('L')

        cdef uint _0, _1, i
        cdef ushort _2
        cdef uchar*vec
        cdef uchar*id_offset
        cdef uint*res
        cdef uint doc_id, dist
        cdef ushort off_set, weight
        cdef int count

        ret = <res_node*> PyMem_Malloc(sizeof(res_node)*top_k)

        for _0 in range(n):
            count = 0
            for _1 in range(self.n_idx):
                clus = bytes_to_uint(clusters)
                vec = self.core[_1][clus]
                for i in range(self.data_num[_1][clus]):
                    # dynamic update result list
                    dist = vec_distance(vec + 8, data, self.n_bytes)
                    if count < top_k:
                        doc_id = bytes_to_uint(vec)
                        off_set = bytes_to_ushort(vec+4)
                        weight = bytes_to_ushort(vec+6)
                        heap_push(ret, dist, doc_id, off_set, weight, count)
                        count += 1
                    elif dist < ret[0].dist:
                        doc_id = bytes_to_uint(vec)
                        off_set = bytes_to_ushort(vec+4)
                        weight = bytes_to_ushort(vec+6)
                        ret[0].dist = dist
                        ret[0].doc_id = doc_id
                        ret[0].off_set = off_set
                        ret[0].weight = weight
                        max_heapify(ret, 0, count)
                    vec += self.n_bytes + 8
                clusters += 4
            for _1 in range(top_k):
                res_dist.append(ret[_1].dist)
                res_docs.append(ret[_1].doc_id)
                res_offset.append(ret[_1].off_set)
                res_weight.append(ret[_1].weight)
                res_idx.append(_0)

            data += self.n_bytes
        PyMem_Free(ret)

        return res_docs, res_offset, res_weight, res_dist, res_idx

    cdef void _save(self, char*save_path):
        cdef FILE*save_file
        cdef uint i, j, num
        save_file = fopen(save_path, "wb")
        fwrite(&self.n_clusters, sizeof(uint), 1, save_file)
        fwrite(&self.n_bytes, sizeof(uint), 1, save_file)
        fwrite(&self.n_idx, sizeof(uint), 1, save_file)
        for i in range(self.n_idx):
            fwrite(self.data_num[i], sizeof(uint), self.n_clusters, save_file)
        for i in range(self.n_idx):
            for j in range(self.n_clusters):
                num = ((self.data_num[i][j] + 1) / Increase + 1)*Increase
                fwrite(self.core[i][j], sizeof(uchar), (self.n_bytes+8)*num, save_file)

    cdef void _load(self, char*load_path):
        cdef FILE*load_file
        cdef uint i, j, num
        load_file = fopen(load_path, "rb")
        fread(&self.n_clusters, sizeof(uint), 1, load_file)
        fread(&self.n_bytes, sizeof(uint), 1, load_file)
        fread(&self.n_idx, sizeof(uint), 1, load_file)
        for i in range(self.n_idx):
            tmp = <uint*> PyMem_Malloc(sizeof(uint)*self.n_clusters)
            fread(tmp, sizeof(uint), self.n_clusters, load_file)
            self.data_num[i] = tmp
        for i in range(self.n_idx):
            for j in range(self.n_clusters):
                num = ((self.data_num[i][j] + 1) / Increase + 1)*Increase
                tmp2 = <uchar*> PyMem_Malloc(sizeof(uchar)*(self.n_bytes+8)*num)
                fread(tmp2, sizeof(uchar), (self.n_bytes+8)*num, load_file)
                self.core[i][j] = tmp2
        self.initialized = 1

    def destroy(self):
        self.free_trie()

    cdef void free_trie(self):
        for i in range(self.n_idx):
            PyMem_Free(self.data_num[i])
        for i in range(self.n_idx):
            for j in range(self.n_clusters):
                PyMem_Free(self.core[i][j])

    def save(self, save_path):
        self._save(bytes(save_path, 'utf8'))

    def load(self, load_path):
        self._load(bytes(load_path, 'utf8'))


# cython: language_level=3, wraparound=False, boundscheck=False
# noinspection PyUnresolvedReferences
# created 2019-03-27

#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# pylint: disable=low-comment-ratio

from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
from cpython cimport array
from libc.stdlib cimport qsort
from libc.stdio cimport fopen, fclose, FILE, fwrite, fread

cdef extern from "limits.h":
    cdef int USHRT_MAX
    cdef unsigned int UINT_MAX

short_max = USHRT_MAX
int_max = UINT_MAX
data_size_per_time = USHRT_MAX
node_size_per_time = USHRT_MAX
DEF data_blocks_increment = 100
DEF node_blocks_increment = 100

DEF search_v_max = 10000
DEF search_c_max = 10000
DEF search_w_max = 10000

ctypedef unsigned int UIDX
ctypedef unsigned short UST
ctypedef unsigned char UCR

cdef struct coordi:
    UST x
    UST y

cdef struct ncoordi:
    UST x
    UIDX y

cdef struct Node:
    coordi left
    coordi right
    coordi child
    coordi parent
    unsigned char key
    coordi _next

cdef struct Data:
    UIDX doc_id
    UST offset
    UST weight
    coordi _next
    ncoordi _neighbor
    coordi parent
    UST num_in
    UST num_out

cdef struct DataDist:
    Data*data
    UST dist
    coordi data_coordi

cdef UIDX bytes_to_label(UCR*all_ids, UST bytes_per_label):
    cdef UIDX label
    cdef UST _
    label = 0
    for _ in range(bytes_per_label):
        label += (256**_) * all_ids[_]
    return label

cdef UST bytes_to_offset(UCR*all_ids):
    cdef UST offset
    offset = all_ids[0] + all_ids[1]*256
    return offset


cdef int cmpfunc(const void*a, const void*b) nogil:
    cdef DataDist*av = <DataDist*>a
    cdef DataDist*bv = <DataDist*>b
    return av.dist - bv.dist

cdef DataDist*sort_Datadist(DataDist*W, UIDX wsize):
    qsort(W, wsize, sizeof(DataDist), &cmpfunc)
    return W

cdef UST vec_distance(UCR*va, UCR*vb, UST bytes_per_vector):
    cdef UST i, dist
    dist = 0

    #for i in prange(bytes_per_vector, num_threads=5):
    for i in range(bytes_per_vector):
        if va[i] != vb[i]:
            dist += 1
    return dist

cdef class IndexCore:
    cdef Node*root_node
    cdef UST bytes_per_vector
    cdef UST bytes_per_label
    cdef UST ef
    cdef UIDX max_insert_iterations
    cdef UIDX max_query_iterations

    cdef Data**all_data
    cdef Node**all_nodes
    cdef coordi**all_neigh
    cdef coordi NULL_coordi

    cdef UIDX nodes_capacity
    cdef UIDX data_capacity
    cdef UIDX max_data_blocks
    cdef UIDX max_node_blocks

    cdef UIDX cur_data_blocks
    cdef UIDX cur_node_blocks

    cdef UIDX num_nodes
    cdef UIDX num_data

    cdef UST num_entry
    cdef UST num_count
    cdef UST max_entry

    cdef UCR*cur_vec
    cdef DataDist*W
    cdef DataDist*C
    cdef Data**V
    cdef coordi*res
    cdef coordi*res_query
    cdef coordi*entry_node
    cdef UCR update_single_neighbor, debugging_mode, initialized

    def __cinit__(self, bytes_per_vector, bytes_per_label, ef,
                  max_insert_iterations=1000,
                  max_query_iterations=5000,
                  debugging_mode=0,
                  update_single_neighbor=1):
        self.bytes_per_vector = bytes_per_vector
        self.bytes_per_label = bytes_per_label
        self.ef = ef
        self.update_single_neighbor = update_single_neighbor
        self.max_insert_iterations = max_insert_iterations
        self.max_query_iterations = max_query_iterations
        self.debugging_mode = debugging_mode

        self.nodes_capacity = 0
        self.data_capacity = 0

        self.max_data_blocks = 0
        self.max_node_blocks = 0

        self.cur_data_blocks = 0
        self.cur_node_blocks = 0

        self.num_nodes = 0
        self.num_data = 0

        self.max_entry = 100
        self.num_entry = 0
        self.num_count = 0

        self.entry_node = <coordi*> PyMem_Malloc(sizeof(coordi)*self.max_entry)
        self.cur_vec = <UCR*> PyMem_Malloc(sizeof(UCR)*self.bytes_per_vector)
        self.V = <Data**> PyMem_Malloc(sizeof(Data*)*search_v_max)
        self.C = <DataDist*> PyMem_Malloc(sizeof(DataDist)*search_c_max)
        self.W = <DataDist*> PyMem_Malloc(sizeof(DataDist)*self.ef*10)
        self.res = <coordi*> PyMem_Malloc(sizeof(coordi)*self.ef)
        self.res_query = <coordi*> PyMem_Malloc(sizeof(coordi)*self.ef*10)

        self.NULL_coordi.x = short_max
        self.NULL_coordi.y = short_max

        self.initialized = 0

    cpdef void index_trie(self, UCR*data, const UIDX num_total, UCR*all_ids, UCR*all_offsets, UCR*all_weights):
        cdef Node*node
        cdef UIDX _0
        cdef UST _1
        cdef UIDX _id
        cdef UST _offset
        cdef coordi parent_coordi

        if not self.initialized:
            self.all_nodes = <Node**> PyMem_Malloc(sizeof(Node*)*node_blocks_increment)
            self.all_data = <Data**> PyMem_Malloc(sizeof(Data*)*data_blocks_increment)
            self.all_neigh = <coordi**> PyMem_Malloc(sizeof(coordi*)*data_blocks_increment)

            block_nodes = <Node*> PyMem_Malloc(sizeof(Node)*node_size_per_time)
            block_data = <Data*> PyMem_Malloc(sizeof(Data)*data_size_per_time)
            block_neigh = <coordi*> PyMem_Malloc(sizeof(coordi)*data_size_per_time*self.ef*2)

            for _0 in range(data_size_per_time*self.ef*2):
                block_neigh[_0].x = self.NULL_coordi.x

            self.all_nodes[0] = block_nodes
            self.all_data[0] = block_data
            self.all_neigh[0] = block_neigh

            self.max_node_blocks = node_blocks_increment
            self.max_data_blocks = data_blocks_increment
            self.cur_node_blocks += 1
            self.cur_data_blocks += 1
            self.nodes_capacity += node_size_per_time
            self.data_capacity += data_size_per_time
            self.root_node = self.id2node(self.new_node(self.NULL_coordi))
            self.initialized = 1

        for _0 in range(num_total):
            node = self.root_node
            parent_coordi.x = 0
            parent_coordi.y = 0
            for _1 in range(self.bytes_per_vector):
                key = data[_1]
                self.cur_vec[_1] = key
                while True:
                    if node.key == 0 or node.key == key:
                        node.key = key
                        if node.child.x == self.NULL_coordi.x:
                            node.child = self.new_node(parent_coordi)
                            parent_coordi = node.child
                        parent_coordi = node.child
                        node = self.id2node(node.child)
                        break

                    elif key < node.key:
                        if node.left.x == self.NULL_coordi.x:
                            node.left = self.new_node(parent_coordi)
                            parent_coordi = node.left
                        parent_coordi = node.left
                        node = self.id2node(node.left)

                    elif key > node.key:
                        if node.right.x == self.NULL_coordi.x:
                            node.right = self.new_node(parent_coordi)
                            parent_coordi = node.right
                        parent_coordi = node.right
                        node = self.id2node(node.right)

            _id = bytes_to_label(all_ids, self.bytes_per_label)
            _offset = bytes_to_offset(all_offsets)
            _weight = bytes_to_offset(all_weights)
            self._index_value(node, parent_coordi, _id, _offset, _weight, self.cur_vec)
            data += self.bytes_per_vector
            all_ids += self.bytes_per_label
            all_offsets += 2
            all_weights += 2

    cdef coordi new_node(self, coordi parent_coordi):
        cdef Node*block_nodes
        cdef UIDX n_block
        cdef UIDX n_position
        cdef Node*node_next
        cdef coordi cur_coordi

        if self.num_nodes >= self.nodes_capacity:
            self.nodes_capacity += node_size_per_time
            block_nodes = <Node*> PyMem_Malloc(sizeof(Node)*node_size_per_time)
            n_block = int(self.nodes_capacity/node_size_per_time)
            # NOTE: judge here if need to update all nodes.
            if n_block > self.max_node_blocks:
                self.all_nodes = <Node**> PyMem_Realloc(
                    self.all_nodes,
                    sizeof(Node*)*(self.max_node_blocks+node_blocks_increment))
                self.max_node_blocks += node_blocks_increment

            # NOTE: then stick to the all nodes pointer list.
            self.all_nodes[n_block - 1] = block_nodes
            self.cur_node_blocks += 1

        cur_coordi.x = int(self.nodes_capacity / node_size_per_time) - 1
        cur_coordi.y = self.num_nodes + node_size_per_time - self.nodes_capacity
        node_next = self.id2node(cur_coordi)
        node_next.left = self.NULL_coordi
        node_next.right = self.NULL_coordi
        node_next.child = self.NULL_coordi
        node_next.key = 0
        node_next._next = self.NULL_coordi
        node_next.parent = parent_coordi
        self.num_nodes += 1
        return cur_coordi

    cdef void _index_value(self, Node*node, coordi parent_coordi, UIDX _id, UST _offset, UST _weight, UCR*cur_vec):
        cdef Data*data_block
        cdef Data*cur_node
        cdef coordi*neigh_block
        cdef coordi cur_coordi
        cdef UIDX n_block, n_position, _i

        # NOTE: upate data blocks and neighbor blocks
        if self.num_data >= self.data_capacity:
            self.data_capacity += data_size_per_time
            data_block = <Data*> PyMem_Malloc(sizeof(Data)*data_size_per_time)
            n_block = int(self.data_capacity / data_size_per_time)
            neigh_block = <coordi*> PyMem_Malloc(
                sizeof(coordi) * data_size_per_time * self.ef * 2)
            # NOTE: initialized all neighbors to NULL
            for _i in range(data_size_per_time*self.ef*2):
                neigh_block[_i] = self.NULL_coordi
            if n_block > self.max_data_blocks:
                self.all_data = <Data**>PyMem_Realloc(
                    self.all_data,
                    sizeof(Data*)*(self.max_data_blocks+data_blocks_increment))
                # NOTE: update neighbor pointers' size
                self.all_neigh = <coordi**>PyMem_Realloc(
                    self.all_neigh,
                    sizeof(coordi*)*(self.max_data_blocks+data_blocks_increment))
                # NOTE: update all_data size
                self.max_data_blocks += data_blocks_increment
            self.all_data[n_block - 1] = data_block
            self.all_neigh[n_block - 1] = neigh_block
            self.cur_data_blocks += 1

        # NOTE: update vector blocks
        n_block = int(self.data_capacity / data_size_per_time) - 1
        n_position = self.num_data + data_size_per_time - self.data_capacity

        cur_node = &self.all_data[n_block][n_position]
        cur_coordi.x = n_block
        cur_coordi.y = n_position
        cur_node.doc_id = _id
        cur_node.offset = _offset
        cur_node.weight = _weight
        cur_node._next.x = self.NULL_coordi.x
        cur_node._neighbor.x = n_block
        cur_node._neighbor.y = n_position * self.ef * 2
        cur_node.num_in = 0
        cur_node.parent = parent_coordi
        # NOTE: update vector in cur_node
        if node._next.x == self.NULL_coordi.x:
            node._next.x = n_block
            node._next.y = n_position
        else:
            new_node = self.id2data(node._next)
            while new_node._next.x != self.NULL_coordi.x:
                new_node = self.id2data(new_node._next)
            new_node._next.x = n_block
            new_node._next.y = n_position
        # NOTE: update cur_node neighbors, entry_points
        self.assign_neighbors(cur_node, cur_coordi)
        self.num_data += 1

    cdef void assign_neighbors(self, Data*cur_node, coordi cur_coordi):
        cdef UIDX _i, _j, _k, count, FLAG
        cdef coordi*res
        cdef Data*candi_node
        cdef UCR is_query
        cdef coordi neigh_coordi
        cdef UCR*_vec
        is_query = 0

        if self.num_data < self.ef:
            if self.num_data == 0:
                cur_node.num_out = 0
            for _i in range(self.num_data):
                self.all_neigh[cur_node._neighbor.x][cur_node._neighbor.y+_i].x = 0
                self.all_neigh[cur_node._neighbor.x][cur_node._neighbor.y+_i].y = _i
                cur_node.num_out = _i + 1
            if self.num_entry < 1:
                self.entry_node[self.num_entry] = cur_coordi
                self.num_entry += 1
        else:
            _vec = self.node2vec(cur_node)
            res = self.search_neighbors(_vec, is_query)
            PyMem_Free(_vec)
            for _i in range(self.ef):
                if res[_i].x != self.NULL_coordi.x:
                    self.all_neigh[cur_node._neighbor.x][cur_node._neighbor.y+_i] = res[_i]
                    cur_node.num_out = _i + 1
                else:
                    break
        # NOTE: add bidirectional connections.
        for _k in range(cur_node.num_out):
            neigh_coordi = self.all_neigh[cur_node._neighbor.x][cur_node._neighbor.y+_k]
            self.update_neighbors(cur_coordi, neigh_coordi)

    cdef coordi*search_neighbors(self, UCR*cur_vec, UCR is_query):
        '''
            V: visited node list
            C: candidate node list
            W: selected node list
            res: return list

        '''
        cdef Data*candi_node, *entry_node
        cdef DataDist c_near, w_far, w_far_new
        cdef UST v_len=0, c_len, w_len
        cdef UIDX i, j, k, max_iterations
        cdef UCR FLAG
        cdef UST eq_dist, max_dist_w, min_dist_c
        cdef coordi*res
        cdef coordi cur_coordi
        cdef UCR*_vec

        # NOTE: get the farest node in W and nearest node in C
        if is_query:
            max_iterations = self.max_query_iterations
        else:
            max_iterations = self.max_insert_iterations
        max_dist_w = 0
        min_dist_c = self.bytes_per_vector + 1
        for i in range(self.num_entry):
            entry_node = self.id2data(self.entry_node[i])
            _vec = self.node2vec(entry_node)
            eq_dist = vec_distance(_vec, cur_vec, self.bytes_per_vector)
            PyMem_Free(_vec)
            self.C[i].dist = eq_dist
            self.C[i].data = entry_node
            self.C[i].data_coordi = self.entry_node[i]
            if i < self.ef * 10:
                self.W[i].dist = eq_dist
                self.W[i].data = entry_node
                self.W[i].data_coordi = self.entry_node[i]

            self.V[i] = entry_node
            v_len += 1

        # v_len = 0
        c_len = min(self.num_entry, self.ef*10)
        w_len = min(self.num_entry, self.ef*10)
        cdef int count
        count = 0

        while c_len > 0:
            max_dist_w = 0
            min_dist_c = self.bytes_per_vector + 1
            for i in range(w_len):
                if self.W[i].dist > max_dist_w:
                    max_dist_w = self.W[i].dist
                    w_far = self.W[i]
            for i in range(c_len):
                if self.C[i].dist < min_dist_c:
                    min_dist_c = self.C[i].dist
                    c_near = self.C[i]
            if (count > max_iterations) or (c_near.dist > w_far.dist):
                self.num_count += 1
                break

            if v_len < search_v_max:
                self.V[v_len] = c_near.data
                v_len += 1

            for i in range(c_near.data.num_out):
                cur_coordi = self.all_neigh[c_near.data._neighbor.x][c_near.data._neighbor.y+i]
                candi_node = self.id2data(cur_coordi)
                FLAG = 1
                # NOTE: check if candi_node in V
                # NOTE: check if candi_node in W already
                for j in range(v_len):
                    if candi_node == self.V[j]:
                        FLAG = 0
                        break
                for j in range(w_len):
                    if candi_node == self.W[j].data:
                        FLAG = 0
                        break
                if FLAG:
                    if v_len < search_v_max:
                        self.V[v_len] = candi_node
                        v_len += 1
                    else:
                        print('visisted list is full ...')
                    _vec = self.node2vec(candi_node)
                    eq_dist = vec_distance(_vec, cur_vec, self.bytes_per_vector)
                    PyMem_Free(_vec)
                    count += 1
                    if w_len < self.ef * 10:
                        self.W[w_len].data = candi_node
                        self.W[w_len].dist = eq_dist
                        self.W[w_len].data_coordi = cur_coordi
                        if max_dist_w < eq_dist:
                            max_dist_w = eq_dist
                            w_far = self.W[w_len]

                        w_len += 1
                        # NOTE: insrt this node into C
                        if c_len < search_c_max:
                            self.C[c_len].data = candi_node
                            self.C[c_len].dist = eq_dist
                            self.C[c_len].data_coordi = cur_coordi
                            c_len += 1

                    elif eq_dist < w_far.dist:
                        # NOTE: insrt this node into W
                        max_dist_w = 0
                        for k in range(w_len):
                            if self.W[k].data == w_far.data:
                                self.W[k].data = candi_node
                                self.W[k].dist = eq_dist
                                self.W[k].data_coordi = cur_coordi
                            if self.W[k].dist > max_dist_w:
                                max_dist_w = self.W[k].dist
                                w_far_new = self.W[k]
                        w_far = w_far_new

                        # NOTE: insrt this node into C
                        if c_len < search_c_max:
                            self.C[c_len].data = candi_node
                            self.C[c_len].dist = eq_dist
                            self.C[c_len].data_coordi = cur_coordi
                            c_len += 1

            # extract c_near
            for i in range(c_len):
                if self.C[i].data == c_near.data:
                    self.C[i] = self.C[c_len-1]
                    break
            c_len -= 1

        sort_Datadist(self.W, w_len)
        if is_query:
            for i in range(self.ef*10):
                if i < w_len:
                    self.res_query[i] = self.W[i].data_coordi
                else:
                    self.res_query[i] = self.NULL_coordi
            res = self.res_query
        else:
            for i in range(self.ef):
                if i < w_len:
                    self.res[i] = self.W[i].data_coordi
                else:
                    self.res[i] = self.NULL_coordi
            res = self.res
        return res

    cdef void update_neighbors(self, coordi cur_coordi, coordi neigh_coordi):
        cdef UST cur_dist, max_dist
        cdef UST i, replace, dist
        cdef Data*org_node, *update_node
        cdef coordi tmp_coordi
        cdef UCR*_vec1, *_vec2
        cdef UST FLAG, _x
        cdef UIDX _y
        org_node = self.id2data(cur_coordi)
        update_node= self.id2data(neigh_coordi)

        max_dist = 0
        FLAG = 1
        if update_node.num_out < self.ef * 2:
            for i in range(update_node.num_out):
                tmp_coordi = self.all_neigh[update_node._neighbor.x][update_node._neighbor.y+i]
                if (tmp_coordi.x == cur_coordi.x) and (tmp_coordi.y == cur_coordi.y):
                    FLAG = 0
                    break
            if FLAG:
                _x = update_node._neighbor.x
                _y = update_node._neighbor.y + update_node.num_out
                self.all_neigh[_x][_y] = cur_coordi
                update_node.num_out += 1
        else:
            _vec1 = self.node2vec(org_node)
            _vec2 = self.node2vec(update_node)
            dist = vec_distance(_vec1, _vec2, self.bytes_per_vector)
            PyMem_Free(_vec1)
            PyMem_Free(_vec2)
            if self.update_single_neighbor:
                max_dist = 0
                for i in range(update_node.num_out):
                    _x = update_node._neighbor.x
                    _y = update_node._neighbor.y + i
                    tmp_coordi = self.all_neigh[_x][_y]
                    _vec1 = self.id2vec(tmp_coordi)
                    _vec2 = self.node2vec(update_node)
                    cur_dist = vec_distance(_vec1, _vec2, self.bytes_per_vector)
                    PyMem_Free(_vec1)
                    PyMem_Free(_vec2)
                    if cur_dist > max_dist:
                        max_dist = cur_dist
                        replace = i
                if max_dist > dist:
                    _x = update_node._neighbor.x
                    _y = update_node._neighbor.y + replace
                    self.all_neigh[_x][_y] = cur_coordi
            else:
                neighbor = <DataDist*> PyMem_Malloc(sizeof(DataDist)*update_node.num_out)
                for i in range(update_node.num_out):
                    _x = update_node._neighbor.x
                    _y = update_node._neighbor.y + i
                    tmp_coordi = self.all_neigh[_x][_y]
                    _vec1 = self.id2vec(tmp_coordi)
                    _vec2 = self.node2vec(update_node)
                    cur_dist = vec_distance(_vec1, _vec2, self.bytes_per_vector)
                    PyMem_Free(_vec1)
                    PyMem_Free(_vec2)
                    neighbor[i].data = self.id2data(tmp_coordi)
                    neighbor[i].dist = cur_dist
                    neighbor[i].data_coordi = tmp_coordi
                sort_Datadist(neighbor, update_node.num_out)
                for i in range(self.ef):
                    _x = update_node._neighbor.x
                    _y = update_node._neighbor.y + i
                    self.all_neigh[_x][_y] = neighbor[i].data_coordi
                if dist < neighbor[self.ef-1].dist:
                    _x = update_node._neighbor.x
                    _y = update_node._neighbor.y + self.ef - 1
                    self.all_neigh[_x][_y] = cur_coordi
                update_node.num_out = self.ef
                PyMem_Free(neighbor)

    cpdef find_batch_trie(self, unsigned char*query, const UIDX num_query):
        cdef array.array final_result = array.array('L')
        cdef array.array final_offset = array.array('L')
        cdef array.array final_weight = array.array('L')
        cdef array.array final_idx = array.array('L')
        cdef Node*node
        cdef Data*dnode
        cdef unsigned char*q_pt
        cdef UIDX _0
        cdef UST _1
        cdef UCR is_match
        q_pt = query

        for _0 in range(num_query):
            node = self.root_node
            is_match = 1
            for _1 in range(self.bytes_per_vector):
                key = q_pt[_1]
                while node:
                    if node.key == key:
                        if node.child.x == self.NULL_coordi.x:
                            is_match = 0
                            break
                        else:
                            node = self.id2node(node.child)
                            break
                    elif key < node.key:
                        if node.left.x == self.NULL_coordi.x:
                            is_match = 0
                            break
                        else:
                            node = self.id2node(node.left)
                    elif key > node.key:
                        if node.right.x == self.NULL_coordi.x:
                            is_match = 0
                            break
                        else:
                            node = self.id2node(node.right)
                if not node:
                    is_match = 0
                if is_match == 0:
                    break
            if is_match == 1 and (node._next.x != self.NULL_coordi.x):
                dnode = self.id2data(node._next)
                final_result.append(dnode.doc_id)
                final_offset.append(dnode.offset)
                final_weight.append(dnode.weight)
                final_idx.append(_0)
                while dnode._next.x != self.NULL_coordi.x:
                    dnode = self.id2data(dnode._next)
                    final_result.append(dnode.doc_id)
                    final_offset.append(dnode.offset)
                    final_weight.append(dnode.weight)
                    final_idx.append(_0)

            q_pt += self.bytes_per_vector
        return final_idx, final_result, final_offset, final_weight

    cpdef nsw_search(self, unsigned char*query, const UIDX num_query, const UIDX top_k):
        cdef array.array res_dist = array.array('L')
        cdef array.array res_docs = array.array('L')
        cdef array.array res_offset = array.array('L')
        cdef array.array res_weight = array.array('L')
        cdef array.array res_idx = array.array('L')

        cdef coordi*res
        cdef UIDX _0
        cdef UST _1
        cdef UCR is_query
        is_query = 1
        for _0 in range(num_query):
            for _1 in range(self.bytes_per_vector):
                self.cur_vec[_1] = query[_1]
            query += self.bytes_per_vector
            res = self.search_neighbors(self.cur_vec, is_query)
            for _1 in range(top_k):
                if res[_1].x != self.NULL_coordi.x:
                    res_idx.append(_0)
                    res_docs.append(self.id2data(res[_1]).doc_id)
                    res_offset.append(self.id2data(res[_1]).offset)
                    res_weight.append(self.id2data(res[_1]).weight)
                    res_dist.append(vec_distance(self.cur_vec, self.id2vec(res[_1]), self.bytes_per_vector))
                else:
                    break

        return res_docs, res_offset, res_weight, res_dist, res_idx

    cpdef force_search(self, unsigned char*query, const UIDX num_query, UIDX top_k):
        cdef array.array res_dist = array.array('L')
        cdef array.array res_docs = array.array('L')
        cdef array.array res_offset = array.array('L')
        cdef array.array res_weight = array.array('L')
        cdef array.array res_idx = array.array('L')
        cdef DataDist w_far
        top_k = min(top_k, self.num_data)

        cdef coordi tmp
        cdef UCR*tmp_vec
        cdef UIDX _0, _1, _2, _3, _id, _dist, _dist_far

        for _0 in range(num_query):
            Q = <DataDist*> PyMem_Malloc(sizeof(DataDist) * top_k)
            w_far.dist = 0
            for _1 in range(self.bytes_per_vector):
                self.cur_vec[_1] = query[_1]
            query += self.bytes_per_vector
            for _2 in range(self.cur_data_blocks):
                for _3 in range(data_size_per_time):
                    _id = _2 * data_size_per_time + _3
                    if _id < self.num_data:
                        tmp.x = _2
                        tmp.y = _3
                        tmp_vec = self.id2vec(tmp)
                        if _id < top_k:
                            Q[_id].data_coordi = tmp
                            Q[_id].dist = vec_distance(self.id2vec(tmp), self.cur_vec, self.bytes_per_vector)
                            Q[_id].data = &self.all_data[_2][_3]
                            if Q[_id].dist > w_far.dist:
                                w_far = Q[_id]
                        else:
                            _dist = vec_distance(self.id2vec(tmp), self.cur_vec, self.bytes_per_vector)
                            if _dist < w_far.dist:
                                _dist_far = 0
                                for _1 in range(top_k):
                                    if Q[_1].data == w_far.data:
                                        Q[_1].data = &self.all_data[_2][_3]
                                        Q[_1].data_coordi = tmp
                                        Q[_1].dist = _dist
                                        break
                                for _1 in range(top_k):
                                    if Q[_1].dist > _dist_far:
                                        _dist_far = Q[_1].dist
                                        w_far = Q[_1]
            sort_Datadist(Q, top_k)
            for _1 in range(top_k):
                res_idx.append(_0)
                res_docs.append(Q[_1].data.doc_id)
                res_offset.append(Q[_1].data.offset)
                res_weight.append(Q[_1].data.weight)
                res_dist.append(Q[_1].dist)
            PyMem_Free(Q)

        return res_docs, res_offset, res_weight, res_dist, res_idx

    def __dealloc__(self):
        self.free_trie()

    cdef Data*id2data(self, coordi data_coordi):
        return &self.all_data[data_coordi.x][data_coordi.y]

    cdef Node*id2node(self, coordi node_coordi):
        return &self.all_nodes[node_coordi.x][node_coordi.y]

    cdef UCR*id2vec(self, coordi data_coordi):
        return self.node2vec(self.id2data(data_coordi))

    cdef UCR*node2vec(self, Data*data):
        cdef Node*node
        cdef Node*node_before
        cdef int i
        cur_vec = <UCR*> PyMem_Malloc(sizeof(UCR)*self.bytes_per_vector)
        node = self.id2node(data.parent)

        i = self.bytes_per_vector-1

        while i >= 0:
            if node.parent.x != self.NULL_coordi.x:
                node_before = node
                node = self.id2node(node_before.parent)
                if self.id2node(node.child) == node_before:
                    cur_vec[i] = node.key
                    i -= 1
            else:
                cur_vec[i] = self.root_node.key
                i -= 1
        return cur_vec

    cdef void _save(self, char*save_path):
        cdef FILE*save_file
        cdef UIDX i
        save_file = fopen(save_path, "wb")
        # NOTE: write cur_data_blocks, cur_node_blocks
        fwrite(&self.cur_data_blocks, sizeof(UIDX), 1, save_file)
        fwrite(&self.cur_node_blocks, sizeof(UIDX), 1, save_file)
        fwrite(&self.num_data, sizeof(UIDX), 1, save_file)
        fwrite(&self.num_nodes, sizeof(UIDX), 1, save_file)
        fwrite(&self.num_entry, sizeof(UST), 1, save_file)
        fwrite(&self.max_data_blocks, sizeof(UIDX), 1, save_file)
        fwrite(&self.max_node_blocks, sizeof(UIDX), 1, save_file)
        fwrite(&self.data_capacity, sizeof(UIDX), 1, save_file)
        fwrite(&self.nodes_capacity, sizeof(UIDX), 1, save_file)
        fwrite(self.entry_node, sizeof(coordi), self.max_entry, save_file)
        for i in range(self.cur_data_blocks):
            fwrite(self.all_data[i], sizeof(Data), data_size_per_time, save_file)
        for i in range(self.cur_node_blocks):
            fwrite(self.all_nodes[i], sizeof(Node), node_size_per_time, save_file)
        for i in range(self.cur_data_blocks):
            fwrite(self.all_neigh[i], sizeof(coordi), data_size_per_time*self.ef*2, save_file)
        fclose(save_file)

    cdef void _load(self, char*load_path):
        cdef FILE*load_file
        cdef UIDX i
        load_file = fopen(load_path, "rb")
        fread(&self.cur_data_blocks, sizeof(UIDX), 1, load_file)
        fread(&self.cur_node_blocks, sizeof(UIDX), 1, load_file)
        fread(&self.num_data, sizeof(UIDX), 1, load_file)
        fread(&self.num_nodes, sizeof(UIDX), 1, load_file)
        fread(&self.num_entry, sizeof(UST), 1, load_file)
        fread(&self.max_data_blocks, sizeof(UIDX), 1, load_file)
        fread(&self.max_node_blocks, sizeof(UIDX), 1, load_file)
        fread(&self.data_capacity, sizeof(UIDX), 1, load_file)
        fread(&self.nodes_capacity, sizeof(UIDX), 1, load_file)
        fread(self.entry_node, sizeof(coordi), self.max_entry, load_file)
        self.all_nodes = <Node**> PyMem_Malloc(sizeof(Node*)*self.max_node_blocks)
        self.all_data = <Data**> PyMem_Malloc(sizeof(Data*)*self.max_data_blocks)
        self.all_neigh = <coordi**> PyMem_Malloc(sizeof(coordi*)*self.max_data_blocks)
        self.initialized = 1

        for i in range(self.cur_data_blocks):
            block_data = <Data*> PyMem_Malloc(sizeof(Data)*data_size_per_time)
            fread(block_data, sizeof(Data), data_size_per_time, load_file)
            self.all_data[i] = block_data
        for i in range(self.cur_node_blocks):
            block_nodes = <Node*> PyMem_Malloc(sizeof(Node)*node_size_per_time)
            fread(block_nodes, sizeof(Node), node_size_per_time, load_file)
            self.all_nodes[i] = block_nodes
        for i in range(self.cur_data_blocks):
            block_neigh = <coordi*> PyMem_Malloc(sizeof(coordi)*data_size_per_time*self.ef*2)
            fread(block_neigh, sizeof(coordi), data_size_per_time*self.ef*2, load_file)
            self.all_neigh[i] = block_neigh

        self.root_node = &self.all_nodes[0][0]
        fclose(load_file)

    def save(self, save_path):
        self._save(bytes(save_path, 'utf8'))

    def load(self, load_path):
        self._load(bytes(load_path, 'utf8'))

    cdef vec_distance(self, UCR*va, UCR*vb):
        cdef UST i, dist
        dist = 0
        for i in range(self.bytes_per_vector):
            if va[i] != vb[i]:
                dist += 1
        return dist

    def debugging(self):
        for i in range(self.num_data-10, self.num_data):
            x = int(i / data_size_per_time)
            y = i - x * data_size_per_time

            print(self.all_data[x][y].doc_id, self.all_neigh[x][y*self.ef*2+1].x, self.all_neigh[x][y*self.ef*2+1].y)

        for i in range(self.num_data-3, self.num_data):
            res = []
            vec = self.node2vec(&self.all_data[0][i])
            for j in range(self.bytes_per_vector):
                res.append(vec[j])
            print(res)
        print('-'*100)
        for i in range(self.num_data-10, self.num_data):
            x = self.all_data[0][i].parent.x
            y = self.all_data[0][i].parent.y
            print(x, y, self.all_nodes[x][y].key)

    def destroy(self):
        self.free_trie()

    cdef void free_trie(self):
        cdef UIDX _
        PyMem_Free(self.V)
        PyMem_Free(self.W)
        PyMem_Free(self.C)
        PyMem_Free(self.res)
        PyMem_Free(self.res_query)
        PyMem_Free(self.entry_node)
        PyMem_Free(self.cur_vec)
        for _ in range(int(self.nodes_capacity/node_size_per_time)):
            PyMem_Free(self.all_nodes[_])
        for _ in range(int(self.data_capacity/data_size_per_time)):
            PyMem_Free(self.all_neigh[_])
        for _ in range(int(self.data_capacity/data_size_per_time)):
            PyMem_Free(self.all_data[_])
        PyMem_Free(self.all_neigh)
        PyMem_Free(self.all_nodes)
        PyMem_Free(self.all_data)


#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import List, Tuple

import numpy as np

from ..indexer.base import BaseKeyIndexer


class DictKeyIndexer(BaseKeyIndexer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._key_info = {}

    def add(self, keys: List[Tuple[int, int]], weights: List[float], *args, **kwargs) -> int:
        for (k, o), w in zip(keys, weights):
            self._key_info[k] = o, w
        return len(self._key_info)

    def query(self, keys: List[int], *args, **kwargs) -> List[Tuple[int, int, float]]:
        return [(k, *self._key_info[k]) for k in keys]

    @property
    def size(self):
        return len(self._key_info)


class ListKeyIndexer(BaseKeyIndexer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._int2key = []  # type: List[Tuple[int, int]]
        self._int2key_weight = []  # type: List[float]

    def add(self, keys: List[Tuple[int, int]], weights: List[float], *args, **kwargs) -> int:
        if len(keys) != len(weights):
            raise ValueError('"keys" and "weights" must have the same length')
        self._int2key.extend(keys)
        self._int2key_weight.extend(weights)
        return len(self._int2key)

    def query(self, keys: List[int], *args, **kwargs) -> List[Tuple[int, int, float]]:
        return [(*self._int2key[k], self._int2key_weight[k]) for k in keys]

    @property
    def size(self):
        return len(self._int2key)


class ListNumpyKeyIndexer(ListKeyIndexer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data_updated = False
        self._np_int2key = None
        self._np_int2key_weight = None

    def _build_np_buffer(self):
        if self._data_updated or not self._np_int2key or not self._np_int2key_weight:
            self._np_int2key = np.array(self._int2key, int)
            self._np_int2key_weight = np.array(self._int2key_weight, float)

    def add(self, *args, **kwargs) -> int:
        self._data_updated = True
        return super().add(*args, **kwargs)

    def query(self, keys: List[int], *args, **kwargs) -> List[Tuple[int, int, float]]:
        self._build_np_buffer()
        key_offset = self._np_int2key[keys, 0:2].astype(int).tolist()
        weights = self._np_int2key_weight[keys].astype(float).tolist()
        return [(*ko, w) for ko, w in zip(key_offset, weights)]

    def __getstate__(self):
        d = super().__getstate__()
        del d['_np_int2key_weight']
        del d['_np_int2key']
        return d


class NumpyKeyIndexer(BaseKeyIndexer):
    def __init__(self, buffer_size: int = 10000, col_size: int = 3, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._int2key_info = np.zeros([buffer_size, col_size])
        self._buffer_size = 10000
        self._col_size = col_size
        self._size = 0
        self._max_size = self._buffer_size

    def add(self, keys: List[Tuple[int, int]], weights: List[float], *args, **kwargs) -> int:
        l = len(keys)
        if self._size + l > self._max_size:
            extend_size = max(l, self._buffer_size)
            self._int2key_info = np.concatenate([self._int2key_info, np.zeros([extend_size, self._col_size])])
            self._max_size += extend_size

        self._int2key_info[self._size:(self._size + l), 0:(self._col_size - 1)] = np.array(keys)
        self._int2key_info[self._size:(self._size + l), self._col_size - 1] = np.array(weights)
        self._size += l
        return self._size

    def query(self, keys: List[int], *args, **kwargs) -> List[Tuple[int, int, float]]:
        key_offset = self._int2key_info[keys, 0:(self._col_size - 1)].astype(int).tolist()
        weights = self._int2key_info[keys, self._col_size - 1].astype(float).tolist()
        return [(*ko, w) for ko, w in zip(key_offset, weights)]

    @property
    def size(self):
        return self._size

    @property
    def capacity(self):
        return self._max_size



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

import numpy as np
import random
from .base import BaseVideoPreprocessor
from ..helper import get_video_frames, phash_descriptor
from ...proto import gnes_pb2, array2blob


class FFmpegPreprocessor(BaseVideoPreprocessor):

    def __init__(self,
                 duplicate_rm: bool = True,
                 use_phash_weight: bool = False,
                 phash_thresh: int = 5,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.phash_thresh = phash_thresh
        self.duplicate_rm = duplicate_rm
        self.use_phash_weight = use_phash_weight

        self._ffmpeg_kwargs = kwargs

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)

        # video could't be processed from ndarray!
        # only bytes can be passed into ffmpeg pipeline
        if doc.raw_bytes:
            frames = get_video_frames(doc.raw_bytes, **self._ffmpeg_kwargs)

            # remove dupliated key frames by phash value
            if self.duplicate_rm:
                frames = self.duplicate_rm_hash(frames)

            if self.use_phash_weight:
                weight = FFmpegPreprocessor.pic_weight(frames)
            else:
                weight = [1 / len(frames)] * len(frames)

            for ci, chunk in enumerate(frames):
                c = doc.chunks.add()
                c.doc_id = doc.doc_id
                c.blob.CopyFrom(array2blob(chunk))
                c.offset_1d = ci
                c.weight = weight[ci]

        else:
            self.logger.error('bad document: "raw_bytes" is empty!')

    @staticmethod
    def pic_weight(images: List['np.ndarray']) -> List[float]:
        import cv2
        weight = np.zeros([len(images)])
        # n_channel is usually 3 for RGB images
        n_channel = images[0].shape[-1]
        for i, image in enumerate(images):
            weight[i] = sum([
                cv2.calcHist([image], [_], None, [256], [0, 256]).var()
                for _ in range(n_channel)
            ])
        weight = weight / weight.sum()

        # normalized result
        weight = np.exp(-weight * 10)
        return weight / weight.sum()

    def duplicate_rm_hash(self,
                          images: List['np.ndarray']) -> List['np.ndarray']:
        hash_list = [phash_descriptor(_) for _ in images]
        ret = []
        for i, h in enumerate(hash_list):
            flag = 1
            if len(ret) >= 1:
                # only keep images with high phash diff
                # comparing only last kept 9 pics
                for j in range(1, min(len(ret) + 1, 9)):
                    dist = abs(ret[-j][1] - h)
                    if dist < self.phash_thresh:
                        flag = 0
                        break

            if flag:
                ret.append((i, h))

        return [images[_[0]] for _ in ret]


class FFmpegVideoSegmentor(BaseVideoPreprocessor):
    def __init__(self,
                 segment_method: str = 'cut_by_frame',
                 segment_interval: int = -1,
                 segment_num: int = 3,
                 max_frames_per_doc: int = -1,
                 audio_interval: int = 30,
                 sample_rate: int = 16000,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.segment_method = segment_method
        self.segment_interval = segment_interval
        self.segment_num = segment_num
        self.max_frames_per_doc = max_frames_per_doc
        self.audio_interval = audio_interval
        self.sample_rate = sample_rate
        self._ffmpeg_kwargs = kwargs

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)
        if doc.raw_bytes:
            frames = get_video_frames(doc.raw_bytes, **self._ffmpeg_kwargs)
            if self.max_frames_per_doc > 0:
                random_id = random.sample(range(len(frames)),
                                          k=min(self.max_frames_per_doc, len(frames)))
                frames = [frames[i] for i in sorted(random_id)]

            sub_videos = []
            if len(frames) >= 1:
                # cut by frame: should specify how many frames to cut
                if self.segment_method == 'cut_by_frame':
                    if self.segment_interval == -1:
                        sub_videos = [frames]
                    else:
                        sub_videos = [frames[_: _ + self.segment_interval]
                                      for _ in range(0, len(frames), self.segment_interval)]
                # cut by num: should specify how many chunks for each doc
                elif self.segment_method == 'cut_by_num':
                    if self.segment_num >= 2:
                        _interval = int(len(frames) / self.segment_num)
                        sub_videos = [frames[_: _ + _interval]
                                      for _ in range(0, len(frames), _interval)]
                    else:
                        sub_videos = [frames]

                for ci, chunk in enumerate(sub_videos):
                    c = doc.chunks.add()
                    c.doc_id = doc.doc_id
                    c.blob.CopyFrom(array2blob(np.array(chunk, dtype=np.uint8)))
                    c.offset_1d = ci
                    c.weight = 1 / len(sub_videos)

            else:
                self.logger.info('bad document: no key frames extracted')
        else:
            self.logger.error('bad document: "raw_bytes" is empty!')



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




from typing import List

import numpy as np

from ..base import BaseTextEncoder
from ...helper import batching, pooling_np


class ElmoEncoder(BaseTextEncoder):
    is_trained = True
    batch_size = 64

    def __init__(self, model_dir: str, pooling_layer: int = -1,
                 pooling_strategy: str = 'REDUCE_MEAN', *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_dir = model_dir

        if pooling_layer > 2:
            raise ValueError('pooling_layer = %d is not supported now!' %
                             pooling_layer)
        self.pooling_layer = pooling_layer
        self.pooling_strategy = pooling_strategy

    def post_init(self):
        from elmoformanylangs import Embedder
        from ...helper import Tokenizer
        self._elmo = Embedder(model_dir=self.model_dir, batch_size=self.batch_size)
        self.cn_tokenizer = Tokenizer()

    @batching
    def encode(self, text: List[str], *args, **kwargs) -> np.ndarray:
        # tokenize text
        batch_tokens = [self.cn_tokenizer.tokenize(sent) for sent in text]

        elmo_encodes = self._elmo.sents2elmo(batch_tokens, output_layer=-2)

        pooled_data = []
        for token_encodes in elmo_encodes:
            if self.pooling_layer == -1:
                _layer_data = np.average(token_encodes, axis=0)
            elif self.pooling_layer >= 0:
                _layer_data = token_encodes[self.pooling_layer]
            else:
                raise ValueError('pooling_layer = %d is not supported now!' %
                                 self.pooling_layer)

            _pooled = pooling_np(_layer_data, self.pooling_strategy)
            pooled_data.append(_pooled)
        return np.array(pooled_data, dtype=np.float32)



import os
import re
from typing import TextIO, List

from gnes.proto import gnes_pb2


def txt_file2pb_docs(fp: TextIO, start_id: int = 0) -> List['gnes_pb2.Document']:
    data = [v for v in fp if v.strip()]
    docs = []
    for doc_id, doc_txt in enumerate(data, start_id):
        doc = line2pb_doc(doc_txt, doc_id)
        docs.append(doc)
    return docs


def line2pb_doc(line: str, doc_id: int = 0, deliminator: str = r'[.。！？!?]+') -> 'gnes_pb2.Document':
    doc = gnes_pb2.Document()
    doc.doc_id = doc_id
    doc.doc_type = gnes_pb2.Document.TEXT
    doc.meta_info = line.encode()
    if deliminator:
        for ci, s in enumerate(re.split(deliminator, line)):
            if s.strip():
                c = doc.chunks.add()
                c.doc_id = doc_id
                c.text = s
                c.offset_1d = ci
    else:
        c = doc.chunks.add()
        c.doc_id = doc_id
        c.text = line
        c.offset_1d = 0
    return doc


env_dict = {
    'orange-ci': {
        'BERT_CI_PORT': 7125,
        'BERT_CI_PORT_OUT': 7126,
        'BERT_CI_MODEL': '/chinese_L-12_H-768_A-12',
        'ELMO_CI_MODEL': '/zhs.model',
        'FLAIR_CI_MODEL': '/multi-forward-fast',
        'GPT_CI_MODEL': '/openai_gpt',
        'GPT2_CI_MODEL': '/openai_gpt2',
        'XL_CI_MODEL': '/transformer_xl_wt103',
        'WORD2VEC_MODEL': '/sgns.wiki.bigram-char.sample',
        'VGG_MODEL': '/',
        'RESNET_MODEL': '/',
        'INCEPTION_MODEL': '/',
        'MOBILENET_MODEL': '/',
        'FASTERRCNN_MODEL': '/',
        'GNES_PROFILING': ''
    },
    'idc-165': {
        'BERT_CI_PORT': 7125,
        'BERT_CI_PORT_OUT': 7126,
        'BERT_CI_MODEL': '/ext_data/chinese_L-12_H-768_A-12',
        'ELMO_CI_MODEL': '/ext_data/zhs.model',
        'FLAIR_CI_MODEL': '/ext_data/multi-forward-fast',
        'GPT_CI_MODEL': '/ext_data/openai_gpt',
        'GPT2_CI_MODEL': '/ext_data/openai_gpt2',
        'XL_CI_MODEL': '/ext_data/transformer_xl_wt103',
        'WORD2VEC_MODEL': '/ext_data/sgns.wiki.bigram-char.sample',
        'VGG_MODEL': '/ext_data/image_encoder',
        'RESNET_MODEL': '/ext_data/image_encoder',
        'INCEPTION_MODEL': '/ext_data/image_encoder',
        'MOBILENET_MODEL': '/ext_data/image_encoder',
        'FASTERRCNN_MODEL': '/ext_data/image_preprocessor',
        'GNES_PROFILING': ''
    }

}

for k, v in env_dict[os.environ.get('GNES_ENV_SET', 'idc-165')].items():
    if k not in os.environ:
        os.environ[k] = str(v)
    else:
        print('os.environ["%s"]=%s exists already, i wont set it to %s' % (k, os.environ[k], str(v)))



import unittest

import numpy as np
from numpy.testing import assert_array_equal

from gnes.encoder.base import BaseEncoder
from gnes.helper import batching, batch_iterator

get_batch_size = lambda x: 2


class bar(BaseEncoder):
    def __init__(self):
        super().__init__()
        self.batch_size = None

    @batching(batch_size=2, num_batch=4)
    def foo(self, data):
        return np.array(data)

    @batching(batch_size=2)
    def bar(self, data):
        return np.array(data)

    @batching(batch_size=8, num_batch=8)
    def foo1(self, data):
        return np.array(data)

    @batching
    def foo2(self, data):
        return np.array(data)

    @batching(batch_size=get_batch_size)
    def foo3(self, data):
        return np.array(data)

    @batching(batch_size=get_batch_size)
    def train(self, data):
        print('train: %s' % data)


class TestBatching(unittest.TestCase):
    def test_iterator(self):
        a = [1, 2, 3, 4, 5, 6, 7]
        b = np.array([[1, 2], [3, 4], [5, 6], [7, 8], [9, 10]])

        num_batches = np.ceil(len(a) / 2)
        act_batch = 0
        for j in batch_iterator(a, 2):
            self.assertLessEqual(len(j), 2)
            self.assertEqual(type(j), list)
            act_batch += 1
        self.assertEqual(num_batches, act_batch)

        num_batches = np.ceil(len(b) / 2)
        act_batch = 0
        for j in batch_iterator(b, 2):
            self.assertLessEqual(j.shape[0], 2)
            self.assertEqual(type(j), np.ndarray)
            act_batch += 1
        self.assertEqual(num_batches, act_batch)

        num_batches = np.ceil(len(a) / 2)
        act_batch = 0
        for j in batch_iterator(iter(a), 2):
            self.assertLessEqual(len(j), 2)
            act_batch += 1
        self.assertEqual(num_batches, act_batch)

    def test_decorator(self):
        b = bar()

        def _test_fn(fn):
            self.assertEqual(fn([1]), np.array([1, ]))
            self.assertSequenceEqual(fn([1, 2]).tolist(), [1, 2])
            self.assertSequenceEqual(fn(list(range(1, 30))).tolist(), list(range(1, 30)))
            self.assertSequenceEqual(fn(range(1, 30)).tolist(), list(range(1, 30)))

        _test_fn(b.foo1)
        _test_fn(b.foo2)
        _test_fn(b.bar)

        self.assertSequenceEqual(b.foo(range(1, 30)).tolist(), list(range(1, 9)))
        t = np.random.randint(0, 255, [32, 10])
        assert_array_equal(b.foo1(t), t)
        assert_array_equal(b.bar(t), t)
        assert_array_equal(b.foo(t), t[:8])

        b.batch_size = 8
        _test_fn(b.foo2)
        _test_fn(b.foo3)
        self.assertEqual(b.train([1]), None)

    def test_mini_batch(self):
        x = list(range(10))

        @batching(batch_size=4)
        def _do_mini_batch(_, y):
            return y

        # this will follow self.batch_size, which is None
        @batching
        def _do_mini_batch2(_, y):
            return y

        self.assertEqual(_do_mini_batch(None, x), [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9]])
        self.assertEqual(_do_mini_batch2(self, x), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

        self.batch_size = 4

        self.assertEqual(_do_mini_batch2(self, x), [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9]])



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.



import os
from typing import List, Tuple, Any

import numpy as np

from .cython import IndexCore
from ...base import BaseVectorIndexer


class BIndexer(BaseVectorIndexer):

    def __init__(self,
                 num_bytes: int = None,
                 ef: int = 20,
                 insert_iterations: int = 200,
                 query_iterations: int = 500,
                 data_path: str = './bindexer_data',
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.num_bytes = num_bytes
        self.ef = ef
        self.insert_iterations = insert_iterations
        self.query_iterations = query_iterations

        self.data_path = data_path
        self._weight_norm = 2 ** 16 - 1

    def post_init(self):
        self.bindexer = IndexCore(self.num_bytes, 4, self.ef,
                                  self.insert_iterations,
                                  self.query_iterations)
        try:
            if not os.path.exists(self.data_path):
                raise FileNotFoundError('"data_path" is not exist')
            if os.path.isdir(self.data_path):
                raise IsADirectoryError('"data_path" must be a file path, not a directory')
            self.bindexer.load(self.data_path)
        except (FileNotFoundError, IsADirectoryError):
            self.logger.warning('fail to load model from %s, will create an empty one' % self.data_path)

    def add(self, keys: List[Tuple[int, Any]], vectors: np.ndarray, weights: List[float], *args,
            **kwargs):
        if len(vectors) != len(keys):
            raise ValueError('vectors length should be equal to doc_ids')

        if vectors.dtype != np.uint8:
            raise ValueError('vectors should be ndarray of uint8')

        num_rows = len(keys)
        keys, offsets = zip(*keys)
        keys = np.array(keys, dtype=np.uint32).tobytes()
        offsets = np.array(offsets, dtype=np.uint16).tobytes()
        weights = np.array([w * self._weight_norm for w in weights], dtype=np.uint16).tobytes()
        self.bindexer.index_trie(vectors.tobytes(), num_rows, keys, offsets, weights)

    def query(self,
              keys: np.ndarray,
              top_k: int,
              method: str = 'nsw',
              *args,
              **kwargs) -> List[List[Tuple]]:

        if keys.dtype != np.uint8:
            raise ValueError("vectors should be ndarray of uint8")

        # num_rows = int(len(keys) / self.num_bytes)
        num_rows = keys.shape[0]
        keys = keys.tobytes()

        result = [[] for _ in range(num_rows)]

        if method == 'nsw':
            # find the indexed items with same value
            q_idx, doc_ids, offsets, weights = self.bindexer.find_batch_trie(
                keys, num_rows)
            for (i, q, o, w) in zip(doc_ids, q_idx, offsets, weights):
                result[q].append((i, o, w / self._weight_norm, 1))

            # search the indexed items with similar value
            doc_ids, offsets, weights, dists, q_idx = self.bindexer.nsw_search(
                keys, num_rows, top_k)
            for (i, o, w, d, q) in zip(doc_ids, offsets, weights, dists, q_idx):
                if d == 0:
                    continue
                result[q].append((i, o, w / self._weight_norm, self.normalize_score(d)))

            # get the top-k
            for q in range(num_rows):
                result[q] = result[q][:top_k]
        elif method == 'force':
            doc_ids, offsets, weights, dists, q_idx = self.bindexer.force_search(
                keys, num_rows, top_k)
            for (i, o, w, d, q) in zip(doc_ids, offsets, weights, dists, q_idx):
                result[q].append((i, o, w / self._weight_norm, self.normalize_score(d)))
        return result

    def normalize_score(self, distance: int, *args, **kwargs) -> float:
        return 1. - distance / self.num_bytes

    def __getstate__(self):
        self.bindexer.save(self.data_path)
        d = super().__getstate__()
        return d



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import zmq

from ..helper import set_logger
from ..proto import send_message, gnes_pb2, recv_message
from ..service.base import build_socket


class ZmqClient:

    def __init__(self, args):
        self.args = args
        self.identity = args.identity if 'identity' in args else None
        self.logger = set_logger(self.__class__.__name__, self.args.verbose)
        self.ctx = zmq.Context()
        self.ctx.setsockopt(zmq.LINGER, 0)
        self.receiver, recv_addr = build_socket(self.ctx, self.args.host_in, self.args.port_in, self.args.socket_in,
                                                getattr(self, 'identity', None))
        self.sender, send_addr = build_socket(self.ctx, self.args.host_out, self.args.port_out, self.args.socket_out,
                                              getattr(self, 'identity', None))
        self.logger.info('send via %s, receive via %s' % (send_addr, recv_addr))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.sender.close()
        self.receiver.close()
        self.ctx.term()

    def send_message(self, message: "gnes_pb2.Message", timeout: int = -1):
        send_message(self.sender, message, timeout=timeout)

    def recv_message(self, timeout: int = -1) -> gnes_pb2.Message:
        return recv_message(self.receiver, timeout=timeout)



import unittest

import numpy as np
from numpy.testing import assert_array_equal

from gnes.proto import gnes_pb2, array2blob, blob2array


class TestProto(unittest.TestCase):

    def test_array_proto(self):
        x = np.random.random([5, 4])
        blob = array2blob(x)
        x1 = blob2array(blob)
        assert_array_equal(x, x1)

    def test_new_msg(self):
        a = gnes_pb2.Message()
        a.response.index.status = gnes_pb2.Response.SUCCESS
        print(a)
        a.request.train.docs.extend([gnes_pb2.Document() for _ in range(2)])
        print(a)
        a.request.train.ClearField('docs')
        a.request.train.docs.extend([gnes_pb2.Document() for _ in range(3)])
        print(a)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


# A key-value map for Class to the (module)file it located in
from ..base import register_all_class

_cls2file_map = {
    'FaissIndexer': 'vector.faiss',
    'LVDBIndexer': 'fulltext.leveldb',
    'AsyncLVDBIndexer': 'fulltext.leveldb',
    'NumpyIndexer': 'vector.numpy',
    'BIndexer': 'vector.bindexer',
    'HBIndexer': 'vector.hbindexer',
    'JointIndexer': 'base',
    'BaseIndexer': 'base',
    'BaseTextIndexer': 'base',
    'AnnoyIndexer': 'vector.annoy'
}

register_all_class(_cls2file_map, 'indexer')



import os
import unittest.mock

import grpc

from gnes.cli.parser import set_router_service_parser, set_frontend_parser
from gnes.proto import gnes_pb2_grpc, RequestGenerator
from gnes.service.base import ServiceManager, SocketType, ParallelType
from gnes.service.frontend import FrontendService
from gnes.service.router import RouterService


class TestServiceManager(unittest.TestCase):
    def setUp(self):
        self.all_bytes = [b'abc', b'def', b'cde'] * 10
        self.all_bytes2 = [b'abc', b'def', b'cde']
        os.unsetenv('http_proxy')
        os.unsetenv('https_proxy')

    def _test_multiple_router(self, backend='thread', num_parallel=5):
        a = set_router_service_parser().parse_args([
            '--yaml_path', 'BaseRouter',
            '--num_parallel', str(num_parallel),
            '--parallel_backend', backend
        ])
        with ServiceManager(RouterService, a):
            pass

    def _test_grpc_multiple_router(self, backend='thread', num_parallel=5):
        args = set_frontend_parser().parse_args([
            '--grpc_host', '127.0.0.1',
        ])

        p_args = set_router_service_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.PULL_CONNECT),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--yaml_path', 'BaseRouter',
            '--num_parallel', str(num_parallel),
            '--parallel_backend', backend
        ])

        with ServiceManager(RouterService, p_args), FrontendService(args), grpc.insecure_channel(
                '%s:%s' % (args.grpc_host, args.grpc_port),
                options=[('grpc.max_send_message_length', 70 * 1024 * 1024),
                         ('grpc.max_receive_message_length', 70 * 1024 * 1024)]) as channel:
            stub = gnes_pb2_grpc.GnesRPCStub(channel)
            resp = stub.Call(list(RequestGenerator.query(b'abc', 1))[0])
            self.assertEqual(resp.request_id, '0')

    def _test_grpc_multiple_pub(self, backend='thread', num_parallel=5):
        args = set_frontend_parser().parse_args([
            '--grpc_host', '127.0.0.1',
        ])

        p_args = set_router_service_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.PULL_CONNECT),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--yaml_path', 'BaseRouter',
            '--num_parallel', str(num_parallel),
            '--parallel_backend', backend,
            '--parallel_type', str(ParallelType.PUB_BLOCK)
        ])

        with ServiceManager(RouterService, p_args), FrontendService(args), grpc.insecure_channel(
                '%s:%s' % (args.grpc_host, args.grpc_port),
                options=[('grpc.max_send_message_length', 70 * 1024 * 1024),
                         ('grpc.max_receive_message_length', 70 * 1024 * 1024)]) as channel:
            stub = gnes_pb2_grpc.GnesRPCStub(channel)
            resp = stub.Call(list(RequestGenerator.query(b'abc', 1))[0])
            self.assertEqual(resp.request_id, '0')

    def test_grpc_with_pub(self):
        self._test_grpc_multiple_pub('thread', 1)
        self._test_grpc_multiple_pub('process', 1)
        self._test_grpc_multiple_pub('thread', 5)
        self._test_grpc_multiple_pub('process', 5)

    def test_grpc_with_multi_service(self):
        self._test_grpc_multiple_router('thread', 1)
        self._test_grpc_multiple_router('process', 1)
        self._test_grpc_multiple_router('thread', 5)
        self._test_grpc_multiple_router('process', 5)

    def test_multiple_router(self):
        self._test_multiple_router('thread', 1)
        self._test_multiple_router('process', 1)
        self._test_multiple_router('thread', 5)
        self._test_multiple_router('process', 5)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

import numpy as np
from PIL import Image

from ..base import BaseImageEncoder
from ...helper import batching, get_first_available_gpu


class TFInceptionEncoder(BaseImageEncoder):
    batch_size = 64

    def __init__(self, model_dir: str,
                 select_layer: str = 'PreLogitsFlatten',
                 use_cuda: bool = False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_dir = model_dir
        self.select_layer = select_layer
        self._use_cuda = use_cuda
        self.inception_size_x = 299
        self.inception_size_y = 299

    def post_init(self):
        import tensorflow as tf
        from .inception_cores.inception_v4 import inception_v4
        from .inception_cores.inception_utils import inception_arg_scope
        import os
        os.environ['CUDA_VISIBLE_DEVICES'] = str(get_first_available_gpu())
        g = tf.Graph()
        with g.as_default():
            arg_scope = inception_arg_scope()
            inception_v4.default_image_size = self.inception_size_x
            self.inputs = tf.placeholder(tf.float32, (None,
                                                      self.inception_size_x,
                                                      self.inception_size_y, 3))

            with tf.contrib.slim.arg_scope(arg_scope):
                self.logits, self.end_points = inception_v4(self.inputs,
                                                            is_training=False,
                                                            dropout_keep_prob=1.0)

            config = tf.ConfigProto(log_device_placement=False)
            if self._use_cuda:
                config.gpu_options.allow_growth = True
            self.sess = tf.Session(config=config)
            self.saver = tf.train.Saver()
            self.saver.restore(self.sess, self.model_dir)

    def encode(self, img: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        img = [(np.array(Image.fromarray(im).resize((self.inception_size_x,
                                                     self.inception_size_y)), dtype=np.float32) * 2 / 255. - 1.) for im
               in img]

        @batching
        def _encode(_, data):
            _, end_points_ = self.sess.run((self.logits, self.end_points),
                                           feed_dict={self.inputs: data})
            return end_points_[self.select_layer]

        return _encode(self, img).astype(np.float32)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




import os
from typing import List, Tuple, Any

import numpy as np

from ..base import BaseVectorIndexer
from ..key_only import ListKeyIndexer


class FaissIndexer(BaseVectorIndexer):

    def __init__(self, num_dim: int, index_key: str, data_path: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_path = data_path
        self.num_dim = num_dim
        self.index_key = index_key
        self._key_info_indexer = ListKeyIndexer()

    def post_init(self):
        import faiss
        try:
            if not os.path.exists(self.data_path):
                raise FileNotFoundError('"data_path" is not exist')
            if os.path.isdir(self.data_path):
                raise IsADirectoryError('"data_path" must be a file path, not a directory')
            self._faiss_index = faiss.read_index(self.data_path)
        except (RuntimeError, FileNotFoundError, IsADirectoryError):
            self.logger.warning('fail to load model from %s, will init an empty one' % self.data_path)
            self._faiss_index = faiss.index_factory(self.num_dim, self.index_key)

    def add(self, keys: List[Tuple[int, Any]], vectors: np.ndarray, weights: List[float], *args, **kwargs):
        if len(vectors) != len(keys):
            raise ValueError("vectors length should be equal to doc_ids")

        if vectors.dtype != np.float32:
            raise ValueError("vectors should be ndarray of float32")

        self._key_info_indexer.add(keys, weights)
        self._faiss_index.add(vectors)

    def query(self, keys: np.ndarray, top_k: int, *args, **kwargs) -> List[List[Tuple]]:
        if keys.dtype != np.float32:
            raise ValueError("vectors should be ndarray of float32")

        score, ids = self._faiss_index.search(keys, top_k)
        score = self.normalize_score(score)
        ret = []
        for _id, _score in zip(ids, score):
            ret_i = []
            chunk_info = self._key_info_indexer.query(_id)
            for c_info, _score_i in zip(chunk_info, _score):
                ret_i.append((*c_info, _score_i))
            ret.append(ret_i)

        return ret

    def normalize_score(self, score: np.ndarray, *args, **kwargs) -> np.ndarray:
        if 'HNSW' in self.index_key:
            return 1 / (1 + np.sqrt(score) / self.num_dim)
        elif 'PQ' or 'Flat' in self.index_key:
            return 1 / (1 + np.abs(np.sqrt(score)))

    @property
    def size(self):
        return self._faiss_index.ntotal

    def __getstate__(self):
        import faiss
        d = super().__getstate__()
        faiss.write_index(self._faiss_index, self.data_path)
        return d



import random
import unittest

import numpy as np

from gnes.helper import get_perm


class TestPartition(unittest.TestCase):

    def test_greedy_partition(self):
        l = 6
        # 6 must be dividable by 2 and 3, otherwise this unit test will fail as there is no
        # perfect partition
        x = np.array([random.random()] * l + [random.random()] * l)
        a = get_perm(x, 2)
        self.assertEqual(sum(x[a[:l]]), sum(x[a[l:]]))

        x = np.array([random.random()] * l + [random.random()] * l + [random.random()] * l)
        b = get_perm(x, 3)
        self.assertEqual(sum(x[b[:l]]), sum(x[b[l:(2 * l)]]))
        self.assertEqual(sum(x[b[:l]]), sum(x[b[(2 * l):]]))



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




import numpy as np

from ..base import BaseNumericEncoder
from ...helper import get_perm, batching, get_optimal_sample_size, train_required


class PCALocalEncoder(BaseNumericEncoder):
    batch_size = 2048

    def __init__(self, output_dim: int, num_locals: int,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert output_dim >= num_locals and output_dim % num_locals == 0, \
            'output_dim should >= num_locals and can be divided by num_locals!'
        self.output_dim = output_dim
        self.num_locals = num_locals
        self.pca_components = None
        self.mean = None

    @batching(batch_size=get_optimal_sample_size, num_batch=1)
    def train(self, vecs: np.ndarray, *args, **kwargs) -> None:
        import faiss
        num_samples, num_dim = vecs.shape
        assert self.output_dim <= num_samples, 'training PCA requires at least %d points, but %d was given' % (
            self.output_dim, num_samples)
        assert self.output_dim < num_dim, 'PCA output dimension should < data dimension, received (%d, %d)' % (
            self.output_dim, num_dim)

        pca = faiss.PCAMatrix(num_dim, self.output_dim)
        self.mean = np.mean(vecs, axis=0)  # 1 x 768
        pca.train(vecs)
        explained_variance_ratio = faiss.vector_to_array(pca.eigenvalues)[:self.output_dim]
        components = faiss.vector_to_array(pca.PCAMat).reshape([-1, num_dim])[:self.output_dim]

        # permutate engive according to variance
        opt_order = get_perm(explained_variance_ratio, self.num_locals)
        comp_tmp = np.reshape(components[opt_order], [self.output_dim, num_dim])

        self.pca_components = np.transpose(comp_tmp)  # 768 x 200

    @train_required
    @batching
    def encode(self, vecs: np.ndarray, *args, **kwargs) -> np.ndarray:
        return np.matmul(vecs - self.mean, self.pca_components)

    def _copy_from(self, x: 'PCALocalEncoder') -> None:
        self.output_dim = x.output_dim
        self.pca_components = x.pca_components
        self.mean = x.mean
        self.num_locals = x.num_locals
        self.is_trained = x.is_trained



import threading
import uuid
from concurrent.futures import ThreadPoolExecutor

import grpc

from ..client.base import ZmqClient
from ..helper import set_logger
from ..proto import gnes_pb2_grpc, gnes_pb2


class FrontendService:

    def __init__(self, args):
        self.logger = set_logger(self.__class__.__name__, args.verbose)
        self.server = grpc.server(
            ThreadPoolExecutor(max_workers=args.max_concurrency),
            options=[('grpc.max_send_message_length', args.max_message_size * 1024 * 1024),
                     ('grpc.max_receive_message_length', args.max_message_size * 1024 * 1024)])
        self.logger.info('start a grpc server with %d workers' % args.max_concurrency)
        gnes_pb2_grpc.add_GnesRPCServicer_to_server(self.GNESServicer(args), self.server)

        self.bind_address = '{0}:{1}'.format(args.grpc_host, args.grpc_port)
        self.server.add_insecure_port(self.bind_address)

    def __enter__(self):
        self.server.start()
        self.logger.info('listening at: %s' % self.bind_address)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.server.stop(None)

    class GNESServicer(gnes_pb2_grpc.GnesRPCServicer):

        def __init__(self, args):
            self.args = args
            self.logger = set_logger(FrontendService.__name__, args.verbose)
            self.zmq_context = self.ZmqContext(args)

        def add_envelope(self, body: 'gnes_pb2.Request', zmq_client: 'ZmqClient'):
            msg = gnes_pb2.Message()
            msg.envelope.client_id = zmq_client.identity if zmq_client.identity else ''
            if body.request_id:
                msg.envelope.request_id = body.request_id
            else:
                msg.envelope.request_id = str(uuid.uuid4())
                self.logger.warning('request_id is missing, filled it with a random uuid!')
            msg.envelope.part_id = 1
            msg.envelope.num_part.append(1)
            msg.envelope.timeout = 5000
            r = msg.envelope.routes.add()
            r.service = FrontendService.__name__
            r.timestamp.GetCurrentTime()
            msg.request.CopyFrom(body)
            return msg

        def remove_envelope(self, m: 'gnes_pb2.Message'):
            resp = m.response
            resp.request_id = m.envelope.request_id
            return resp

        def Call(self, request, context):
            self.logger.info('received a new request: %s' % request.request_id or 'EMPTY_REQUEST_ID')
            with self.zmq_context as zmq_client:
                zmq_client.send_message(self.add_envelope(request, zmq_client), self.args.timeout)
                return self.remove_envelope(zmq_client.recv_message(self.args.timeout))

        def Train(self, request, context):
            return self.Call(request, context)

        def Index(self, request, context):
            return self.Call(request, context)

        def Search(self, request, context):
            return self.Call(request, context)

        def StreamCall(self, request_iterator, context):
            num_result = 0
            with self.zmq_context as zmq_client:
                for request in request_iterator:
                    zmq_client.send_message(self.add_envelope(request, zmq_client), self.args.timeout)
                    num_result += 1
                for _ in range(num_result):
                    yield self.remove_envelope(zmq_client.recv_message(self.args.timeout))

        class ZmqContext:
            """The zmq context class."""

            def __init__(self, args):
                self.args = args
                self.tlocal = threading.local()
                self.tlocal.client = None

            def __enter__(self):
                """Enter the context."""
                client = ZmqClient(self.args)
                self.tlocal.client = client
                return client

            def __exit__(self, exc_type, exc_value, exc_traceback):
                """Exit the context."""
                self.tlocal.client.close()
                self.tlocal.client = None



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import sys

from termcolor import colored

from . import api
from .parser import get_main_parser

__all__ = ['main']


def get_run_args(parser_fn=get_main_parser, printed=True):
    parser = parser_fn()
    if len(sys.argv) > 1:
        args = parser.parse_args()
        if printed:
            param_str = '\n'.join(['%20s = %s' % (colored(k, 'yellow'), v) for k, v in sorted(vars(args).items())])
            print('usage: %s\n%s\n%s\n' % (' '.join(sys.argv), '_' * 50, param_str))
        return args
    else:
        parser.print_help()
        exit()


def main():
    args = get_run_args()
    getattr(api, args.cli, no_cli_error)(args)


def no_cli_error(*args, **kwargs):
    get_main_parser().print_help()
    exit()



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

import numpy as np

from ..base import BaseReduceRouter
from ...proto import gnes_pb2, blob2array, array2blob


class ConcatEmbedRouter(BaseReduceRouter):

    def apply(self, msg: 'gnes_pb2.Message', accum_msgs: List['gnes_pb2.Message'], *args, **kwargs):
        body = getattr(msg, msg.WhichOneof('body'))
        msg_type = type(getattr(body, body.WhichOneof('body')))
        if msg_type == gnes_pb2.Request.QueryRequest:
            msg.request.search.query.chunk_embeddings.CopyFrom(array2blob(
                np.concatenate([blob2array(m.request.search.query.chunk_embeddings) for m in accum_msgs], axis=1)))
        elif msg_type == gnes_pb2.Request.IndexRequest:
            for i in range(len(msg.request.index.docs)):
                msg.request.index.docs[i].chunk_embeddings.CopyFrom(array2blob(
                    np.concatenate([blob2array(m.request.index.docs[i].chunk_embeddings) for m in accum_msgs], axis=1)))
        else:
            self.logger.error('dont know how to handle %s' % msg_type)

        super().apply(msg, accum_msgs)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List
from collections import defaultdict

from ..base import BaseReduceRouter
from ...proto import gnes_pb2


class ChunkSumRouter(BaseReduceRouter):
    def apply(self, msg: 'gnes_pb2.Message', accum_msgs: List['gnes_pb2.Message'], *args, **kwargs):
        all_docs_id = [top_k.doc.doc_id for m in accum_msgs for top_k in m.response.search.topk_results]
        all_docs_socres = [top_k.score for m in accum_msgs for top_k in m.response.search.topk_results]
        all_score_explained = [top_k.score_explained for m in accum_msgs for top_k in m.response.search.topk_results]

        doc_score = defaultdict(float)
        doc_score_explained = defaultdict(str)

        for id_, s, ex in zip(all_docs_id, all_docs_socres, all_score_explained):
            doc_score[id_] += s
            doc_score_explained[id_] += '%s\n' % ex

        msg.response.search.ClearField('topk_results')

        # sort and add docs
        for k, v in sorted(doc_score.items(), key=lambda kv: -kv[1]):
            r = msg.response.search.topk_results.add()
            r.doc.doc_id = k
            r.score = v
            r.score_explained = doc_score_explained[k]

        super().apply(msg, accum_msgs)


#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import numpy as np

from .base import BaseService as BS, MessageHandler, ServiceError
from ..proto import gnes_pb2, blob2array


class IndexerService(BS):
    handler = MessageHandler(BS.handler)

    def post_init(self):
        from ..indexer.base import BaseIndexer
        self._model = self.load_model(BaseIndexer)

    @handler.register(gnes_pb2.Request.IndexRequest)
    def _handler_index(self, msg: 'gnes_pb2.Message'):
        all_vecs = []
        doc_ids = []
        offsets = []
        weights = []

        for d in msg.request.index.docs:
            if len(d.chunks):
                all_vecs.append(blob2array(d.chunk_embeddings))
                doc_ids += [d.doc_id] * len(d.chunks)
                if d.doc_type == gnes_pb2.Document.TEXT:
                    offsets += [c.offset_1d for c in d.chunks]
                elif d.doc_type == gnes_pb2.Document.IMAGE:
                    offsets += [c.offset_nd for c in d.chunks]
                elif d.doc_type == gnes_pb2.Document.VIDEO:
                    offsets += [c.offset_1d for c in d.chunks]
                elif d.doc_type == gnes_pb2.Document.AUDIO:
                    offsets += [c.offset_1d for c in d.chunks]
                weights += [c.weight for c in d.chunks]

        from ..indexer.base import BaseVectorIndexer, BaseTextIndexer
        if isinstance(self._model, BaseVectorIndexer):
            self._model.add(list(zip(doc_ids, offsets)),
                            np.concatenate(all_vecs, 0),
                            weights)
        elif isinstance(self._model, BaseTextIndexer):
            self._model.add([d.doc_id for d in msg.request.index.docs],
                            [d for d in msg.request.index.docs],
                            [d.weight for d in msg.request.index.docs])

        msg.response.index.status = gnes_pb2.Response.SUCCESS
        self.is_model_changed.set()

    @handler.register(gnes_pb2.Request.QueryRequest)
    def _handler_chunk_search(self, msg: 'gnes_pb2.Message'):
        def _cal_offset_relevance(q_offset, i_offset):
            import math
            if not isinstance(q_offset, list) and isinstance(i_offset, list):
                raise TypeError("Type of qc_offset and offset is supposed to be (list, list), "
                                "but actually we got (%s, %s)" % (str(type(q_offset)), str(type(i_offset))))
            if not len(q_offset) == 2 and len(i_offset) == 2:
                raise ValueError("Length of qc_offset and offset should be (2, 2), "
                                 "but actually we got (%d, %d)" % (len(q_offset), len(i_offset)))
            return 1 / (1 + math.sqrt((q_offset[0] - i_offset[0]) ** 2 + (q_offset[1] - i_offset[1]) ** 2))

        vecs = blob2array(msg.request.search.query.chunk_embeddings)
        q_offset = [c.offset_nd if msg.request.search.query.doc_type == 'IMAGE'
                    else c.offset_1d for c in msg.request.search.query.chunks]
        topk = msg.request.search.top_k
        results = self._model.query(vecs, top_k=msg.request.search.top_k)
        q_weights = [qc.weight for qc in msg.request.search.query.chunks]
        for all_topks, qc_weight, qc_offset in zip(results, q_weights, q_offset):
            for _doc_id, _offset, _weight, _relevance in all_topks:
                r = msg.response.search.topk_results.add()
                r.chunk.doc_id = _doc_id
                r.chunk.weight = _weight
                if msg.request.search.query.doc_type == 'IMAGE':
                    r.chunk.offset_nd = _offset
                    offset_relevance = _cal_offset_relevance(qc_offset, _offset)
                else:
                    r.chunk.offset_1d = _offset
                    offset_relevance = 1
                r.score = _weight * qc_weight * _relevance * offset_relevance
                r.score_explained = '[chunk_score at doc: %d, offset: %d] = ' \
                                    '(doc_chunk_weight: %.6f) * ' \
                                    '(query_doc_chunk_relevance: %.6f) * ' \
                                    '(query_doc_offset_relevance: %.6f) * ' \
                                    '(query_chunk_weight: %.6f)' % (
                                        _doc_id, _offset, _weight, _relevance, offset_relevance, qc_weight)
            msg.response.search.top_k = topk

    @handler.register(gnes_pb2.Response.QueryResponse)
    def _handler_doc_search(self, msg: 'gnes_pb2.Message'):
        if msg.response.search.level == gnes_pb2.Response.QueryResponse.DOCUMENT_NOT_FILLED:
            doc_ids = [r.doc.doc_id for r in msg.response.search.topk_results]
            docs = self._model.query(doc_ids)
            for r, d in zip(msg.response.search.topk_results, docs):
                if d is not None:
                    # fill in the doc if this shard returns non-empty
                    r.doc.CopyFrom(d)
                    r.score *= d.weight
                    r.score_explained = '{%s} * [doc: %d] (doc_weight: %.6f)' % (
                        r.score_explained, d.doc_id, d.weight)
            msg.response.search.level = gnes_pb2.Response.QueryResponse.DOCUMENT
        else:
            raise ServiceError('i dont know how to handle QueryResponse with %s level' % msg.response.search.level)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import io

from .base import YamlComposer
from ..cli.parser import set_composer_parser
from ..helper import set_logger


class YamlComposerFlask:
    def __init__(self, args):
        self.args = args
        self.logger = set_logger(self.__class__.__name__, self.args.verbose)

    def _create_flask_app(self):
        try:
            from flask import Flask, request
            from flask_compress import Compress
        except ImportError:
            raise ImportError('Flask or its dependencies are not fully installed, '
                              'they are required for serving HTTP requests.'
                              'Please use "pip install gnes[http]" to install it.')

        app = Flask(__name__)
        args = set_composer_parser().parse_args([])
        default_html = YamlComposer(args).build_all()['html']

        @app.errorhandler(500)
        def exception_handler(error):
            self.logger.error('unhandled error, i better quit and restart! %s' % error)
            return '<h1>500 Internal Error</h1> ' \
                   'While we are fixing the issue, do you know you can deploy GNES board locally on your machine? ' \
                   'Simply run <pre>docker run -d -p 0.0.0.0:80:8080/tcp gnes/gnes compose --flask</pre>', 500

        @app.route('/', methods=['GET'])
        def _get_homepage():
            return default_html

        @app.route('/generate', methods=['POST'])
        def _regenerate():
            data = request.form if request.form else request.json
            if not data or 'yaml-config' not in data:
                return '<h1>Bad POST request</h1> your POST request does not contain "yaml-config" field!', 406
            try:
                args.yaml_path = io.StringIO(data['yaml-config'])
                if data.get('mermaid_direction', 'top-down').lower() == 'left-right':
                    args.mermaid_leftright = True
                else:
                    args.mermaid_leftright = False
                if 'docker-image' in data:
                    args.docker_img = data['docker-image']
                else:
                    args.docker_img = 'gnes/gnes:latest'

                return YamlComposer(args).build_all()['html']
            except Exception as e:
                self.logger.error(e)
                return '<h1>Bad YAML input</h1> please kindly check the format, indent and content of your YAML file!', 400

        Compress().init_app(app)
        return app

    def run(self):
        app = self._create_flask_app()
        app.run(port=self.args.http_port, threaded=True, host='0.0.0.0')



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import numpy as np

from .base import BaseVideoPreprocessor
from ..helper import get_video_frames, compute_descriptor, compare_descriptor
from ...proto import gnes_pb2, array2blob


class ShotDetectPreprocessor(BaseVideoPreprocessor):
    store_args_kwargs = True

    def __init__(self,
                 frame_size: str = '192*168',
                 descriptor: str = 'block_hsv_histogram',
                 distance_metric: str = 'bhattacharya',
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.frame_size = frame_size
        self.descriptor = descriptor
        self.distance_metric = distance_metric
        self._detector_kwargs = kwargs

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)
        from sklearn.cluster import KMeans

        if doc.raw_bytes:
            # stream_data = io.BytesIO(doc.raw_bytes)
            # vidcap = cv2.VideoCapture(stream_data)
            frames = get_video_frames(
                doc.raw_bytes,
                s=self.frame_size,
                vsync='vfr',
                vf='select=eq(pict_type\\,I)')

            descriptors = []
            shots = []
            for frame in frames:
                descriptor = compute_descriptor(
                    frame, method=self.descriptor, **self._detector_kwargs)
                descriptors.append(descriptor)

            # compute distances between frames
            dists = [
                compare_descriptor(pair[0], pair[1], self.distance_metric)
                for pair in zip(descriptors[:-1], descriptors[1:])
            ]

            dists = np.array(dists).reshape([-1, 1])
            clt = KMeans(n_clusters=2)
            clt.fit(dists)

            # select which cluster includes shot frames
            big_center = np.argmax(clt.cluster_centers_)

            shots = []
            prev_shot = 0
            for i in range(0, len(clt.labels_)):
                if big_center == clt.labels_[i]:
                    shots.append((prev_shot, i + 2))
                    prev_shot = i + 2

            for ci, (start, end) in enumerate(shots):
                c = doc.chunks.add()
                c.doc_id = doc.doc_id
                chunk_pos = start + (end - start) // 2
                chunk = frames[chunk_pos]
                c.blob.CopyFrom(array2blob(chunk))
                c.offset_1d = ci
                c.weight = (end - start) / len(frames)

        else:
            self.logger.error('bad document: "raw_bytes" is empty!')



import os
import unittest

import numpy as np

from gnes.indexer.base import JointIndexer
from tests import txt_file2pb_docs


@unittest.SkipTest
class TestJointIndexer(unittest.TestCase):

    def setUp(self):
        dirname = os.path.dirname(__file__)
        self.dump_path = os.path.join(dirname, 'indexer.bin')
        self.yaml_path = os.path.join(dirname, 'yaml', 'base-indexer.yml')

        self.n_bytes = 8
        self.query_num = 3

        self.docs = []
        self.querys = []

        self.pb_docs = txt_file2pb_docs(open(os.path.join(dirname, 'tangshi.txt')))

    def test_add(self):
        mhi = JointIndexer.load_yaml(self.yaml_path)

        for doc in self.pb_docs:
            if len(doc.chunks) == 0:
                continue
            mhi.add([doc.doc_id], [doc], [1.])
            vecs = np.random.randint(
                0, 255, [len(doc.chunks), self.n_bytes]).astype(np.uint8)
            mhi.add([(doc.doc_id, j) for j in range(len(doc.chunks))], vecs, [1.] * len(doc.chunks))
            if len(self.querys) < self.query_num:
                self.querys.append(vecs)

        for q in self.querys:
            results = mhi.query(q, top_k=1)
            for topk in results:
                print(topk)
                d, o, w, s, *_ = topk[0]
                self.assertEqual(1.0, s)
        mhi.close()



import os
import unittest
from shutil import rmtree

from gnes.base import TrainableType
from gnes.encoder.base import PipelineEncoder
from gnes.encoder.numeric.pca import PCALocalEncoder
from gnes.encoder.numeric.pq import PQEncoder
from gnes.encoder.numeric.tf_pq import TFPQEncoder


class foo(metaclass=TrainableType):
    def __init__(self, *args, **kwargs):
        pass

    def close(self):
        pass


class foo1(foo):
    store_args_kwargs = False

    def __init__(self, a, b=1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        pass


class foo2(foo1):
    def __init__(self, c, *args, **kwargs):
        super().__init__(*args, **kwargs)
        pass


class dummyPipeline(PipelineEncoder):
    store_args_kwargs = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.component = lambda: [foo1(*args, **kwargs),
                                  foo2(*args, **kwargs), ]


class TestYaml(unittest.TestCase):
    def setUp(self):
        dirname = os.path.dirname(__file__)
        self.dump_path = os.path.join(dirname, 'dump.yml')
        self.db_path = './test_leveldb'

    def tearDown(self):
        if os.path.exists(self.db_path):
            rmtree(self.db_path)
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)

    def test_signature(self):
        a = foo1(2)
        self.assertEqual(a._init_kwargs_dict, {'a': 2, 'b': 1})
        a = foo2(2, 3, 3)
        self.assertEqual(a._init_kwargs_dict, {'c': 2, 'a': 3, 'b': 3})
        self.assertRaises(TypeError, foo2, 2, 3, c=2)
        a = foo2(2, 3, 4, 5, 6, 7)
        self.assertEqual(a._init_kwargs_dict, {'c': 2, 'a': 3, 'b': 4})
        a = foo2(2, 3, wee=4)
        self.assertEqual(a._init_kwargs_dict, {'c': 2, 'a': 3, 'b': 1})
        a = foo2(b=1, wee=4, a=3, c=2)
        self.assertEqual(a._init_kwargs_dict, {'c': 2, 'a': 3, 'b': 1})

    def test_dump(self):
        pe0 = dummyPipeline(a=23, b=32, c=['123', '456'])
        pe0.dump_yaml(self.dump_path)
        self.assertTrue(os.path.exists(self.dump_path))
        pe = dummyPipeline.load_yaml(self.dump_path)
        self.assertEqual(type(pe), dummyPipeline)
        self.assertEqual(pe0._init_kwargs_dict, pe._init_kwargs_dict)

    def test_dump2(self):
        dummyPipeline.store_args_kwargs = False
        pe0 = dummyPipeline(a=23, b=32, c=['123', '456'])
        self.assertEqual(pe0._init_kwargs_dict, {})
        dummyPipeline.store_args_kwargs = True
        pe0 = dummyPipeline(a=23, b=32, c=['123', '456'])
        self.assertEqual(pe0._init_kwargs_dict, {'kwargs': dict(a=23, b=32, c=['123', '456'])})
        pe0.dump_yaml(self.dump_path)
        self.assertTrue(os.path.exists(self.dump_path))
        dummyPipeline.store_args_kwargs = False
        pe = dummyPipeline.load_yaml(self.dump_path)
        self.assertEqual(type(pe), dummyPipeline)
        self.assertEqual(pe._init_kwargs_dict, {})
        dummyPipeline.store_args_kwargs = True

    def test_load(self):
        with open(self.dump_path, 'w') as fp:
            fp.write("!PipelineEncoder\n\
                                parameter:\n\
                                  kwargs:\n\
                                    a: 23\n\
                                    b: '32'\n\
                                    c: ['123', '456']")
        PipelineEncoder.store_args_kwargs = True
        pe = dummyPipeline.load_yaml(self.dump_path)
        self.assertEqual(type(pe), PipelineEncoder)
        self.assertEqual(pe._init_kwargs_dict, {'kwargs': dict(a=23, b=32, c=['123', '456'])})
        PipelineEncoder.store_args_kwargs = False
        pe = dummyPipeline.load_yaml(self.dump_path)
        self.assertEqual(type(pe), PipelineEncoder)
        self.assertEqual(pe._init_kwargs_dict, {})

        with open(self.dump_path, 'w') as fp:
            fp.write("!PipelineEncoder\n\
                                parameter:\n\
                                  args:\n\
                                    - 23\n\
                                    - '32'\n\
                                    - ['123', '456']")
        PipelineEncoder.store_args_kwargs = True
        pe = dummyPipeline.load_yaml(self.dump_path)
        self.assertEqual(type(pe), PipelineEncoder)
        self.assertEqual(pe._init_kwargs_dict, {'args': (23, 32, ['123', '456'])})

        PipelineEncoder.store_args_kwargs = False
        pe = dummyPipeline.load_yaml(self.dump_path)
        self.assertEqual(type(pe), PipelineEncoder)
        self.assertEqual(pe._init_kwargs_dict, {})

    def test_nest_pipeline(self):
        self._test_different_encoder_yamlize(dummyPipeline, a=1, b=2, c=3, wee=4)
        self._test_different_encoder_yamlize(PQEncoder, 10)
        self._test_different_encoder_yamlize(TFPQEncoder, 10)
        self._test_different_encoder_yamlize(PCALocalEncoder, 20, 10)

    def _test_different_encoder_yamlize(self, cls, *args, **kwargs):
        a = cls(*args, **kwargs)
        a.dump_yaml(self.dump_path)
        a.close()
        self.assertTrue(os.path.exists(self.dump_path))
        b = cls.load_yaml(self.dump_path)
        self.assertEqual(type(b), cls)
        self.assertEqual(a._init_kwargs_dict, b._init_kwargs_dict)
        b.close()

    @unittest.SkipTest
    def test_NES_yaml_dump(self):
        self._test_different_encoder_yamlize(GNES, num_bytes=8,
                                             pca_output_dim=32,
                                             cluster_per_byte=8,
                                             port=1,
                                             port_out=2,
                                             data_path=self.db_path,
                                             ignore_all_checks=True)

    @unittest.SkipTest
    def test_double_dump(self):
        a = GNES(num_bytes=8,
                 pca_output_dim=32,
                 cluster_per_byte=8,
                 port=1,
                 port_out=2,
                 data_path=self.db_path,
                 ignore_all_checks=True)
        a.dump_yaml(self.dump_path)
        a.close()
        with open(self.dump_path) as fp:
            content_a = fp.readlines()
        b = GNES.load_yaml(self.dump_path)
        b.dump_yaml(self.dump_path)
        b.close()
        with open(self.dump_path) as fp:
            content_b = fp.readlines()
        self.assertEqual(content_a, content_b)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import List, Any

import numpy as np

from ..base import TrainableBase, CompositionalTrainableBase


class BaseEncoder(TrainableBase):

    def encode(self, data: Any, *args, **kwargs) -> Any:
        pass

    def _copy_from(self, x: 'BaseEncoder') -> None:
        pass


class BaseImageEncoder(BaseEncoder):

    def encode(self, img: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        pass


class BaseVideoEncoder(BaseEncoder):

    def encode(self, data: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        pass


class BaseTextEncoder(BaseEncoder):

    def encode(self, text: List[str], *args, **kwargs) -> np.ndarray:
        pass


class BaseNumericEncoder(BaseEncoder):

    def encode(self, text: np.ndarray, *args, **kwargs) -> np.ndarray:
        pass


class BaseAudioEncoder(BaseEncoder):

    def encode(self, data: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        pass


class BaseBinaryEncoder(BaseEncoder):

    def encode(self, data: np.ndarray, *args, **kwargs) -> bytes:
        if data.dtype != np.uint8:
            raise ValueError('data must be np.uint8 but received %s' % data.dtype)
        return data.tobytes()


class PipelineEncoder(CompositionalTrainableBase):
    def encode(self, data: Any, *args, **kwargs) -> Any:
        if not self.component:
            raise NotImplementedError
        for be in self.component:
            data = be.encode(data, *args, **kwargs)
        return data

    def train(self, data, *args, **kwargs):
        if not self.component:
            raise NotImplementedError
        for idx, be in enumerate(self.component):
            be.train(data, *args, **kwargs)
            if idx + 1 < len(self.component):
                data = be.encode(data, *args, **kwargs)



import os
import unittest

from gnes.cli.parser import set_composer_parser, set_composer_flask_parser
from gnes.composer.base import YamlComposer
from gnes.composer.flask import YamlComposerFlask


class TestCompose(unittest.TestCase):
    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.html_path = os.path.join(self.dirname, 'test.html')

    def test_all(self):
        paths = [os.path.join(self.dirname, 'yaml', 'topology%d.yml' % j) for j in range(1, 8)]
        b_a = [(3, 3), (4, 4), (4, 5), (4, 7), (4, 6), (4, 8), (4, 8)]
        for p, j in zip(paths, b_a):
            self._test_topology(p, *j)

    def _test_topology(self, yaml_path: str, num_layer_before: int, num_layer_after: int):
        args = set_composer_parser().parse_args([
            '--yaml_path', yaml_path,
            '--html_path', self.html_path
        ])
        a = YamlComposer(args)
        self.assertEqual(len(a._layers), num_layer_before)
        r = a.build_layers()
        self.assertEqual(len(r), num_layer_after)
        for c in r:
            print(c)
        a.build_all()
        print(a.build_shell(r))
        os.path.exists(self.html_path)
        print(a.build_dockerswarm(r))

    @unittest.SkipTest
    def test_flask_local(self):
        args = set_composer_flask_parser().parse_args(['--flask'])
        YamlComposerFlask(args).run()

    def test_flask(self):
        yaml_path = os.path.join(self.dirname, 'yaml', 'topology1.yml')
        args = set_composer_flask_parser().parse_args([
            '--flask',
            '--yaml_path', yaml_path,
            '--html_path', self.html_path
        ])
        app = YamlComposerFlask(args)._create_flask_app().test_client()
        response = app.get('/', follow_redirects=True)
        self.assertEqual(response.status_code, 200)

        response = app.post('/generate', follow_redirects=True)
        self.assertEqual(response.status_code, 406)

        response = app.post('/generate', data={'yaml-config': ''},
                            follow_redirects=True)
        self.assertEqual(response.status_code, 400)

        response = app.post('/generate',
                            data={'yaml-config': 'port: 5566\nservices:\n- name: Preprocessor\n- name: Encoder'},
                            follow_redirects=True)
        self.assertEqual(response.status_code, 200)

        response = app.get('/', follow_redirects=True)
        self.assertEqual(response.status_code, 200)

    def tearDown(self):
        if os.path.exists(self.html_path):
            os.remove(self.html_path)



import os
import unittest

from gnes.encoder.base import BaseEncoder, PipelineEncoder


class DummyTFEncoder(BaseEncoder):
    is_trained = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def post_init(self):
        import tensorflow as tf
        with tf.Graph().as_default():
            self.a = tf.get_variable(name='a', shape=[])
            self.sess = tf.Session()

    def encode(self, a, *args):
        return self.sess.run(self.a + 1, feed_dict={self.a: a})


class TestLoadDumpPipeline(unittest.TestCase):
    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.yaml_path = os.path.join(self.dirname, 'yaml', 'dummy-pipeline.yml')
        self.dump_path = os.path.join(self.dirname, 'dummy-pipeline.bin')

    def test_base(self):
        a = BaseEncoder.load_yaml(self.yaml_path)
        self.assertFalse(a.is_trained)
        # simulate training
        a.is_trained = True
        a.dump()
        os.path.exists(self.dump_path)

        # load the dump from yaml
        b = BaseEncoder.load_yaml(self.yaml_path)
        self.assertTrue(b.is_trained)

    def test_name_warning(self):
        d1 = DummyTFEncoder()
        d2 = DummyTFEncoder()
        d1.name = ''
        d2.name = ''
        d3 = PipelineEncoder()
        d3.component = lambda: [d1, d2]
        d3.name = 'dummy-pipeline'
        d3.work_dir = './'
        d3.dump()
        d3.dump_yaml()
        print('there should not be any warning after this line')
        BaseEncoder.load_yaml(d3.yaml_full_path)

    def test_dummytf(self):
        d1 = DummyTFEncoder()
        self.assertEqual(d1.encode(1), 2)
        self.assertTrue(d1.is_trained)
        d1.dump()
        d11 = BaseEncoder.load(d1.dump_full_path)
        self.assertTrue(d11.is_trained)

        d2 = DummyTFEncoder()
        self.assertEqual(d2.encode(2), 3)
        self.assertTrue(d2.is_trained)

        d3 = PipelineEncoder()
        d3.component = lambda: [d1, d2]
        self.assertEqual(d3.encode(1), 3)
        self.assertFalse(d3.is_trained)
        self.assertTrue(d3.component[0].is_trained)
        self.assertTrue(d3.component[1].is_trained)

        d3.dump()
        d31 = BaseEncoder.load(d3.dump_full_path)
        self.assertFalse(d31.is_trained)
        self.assertTrue(d31.component[0].is_trained)
        self.assertTrue(d31.component[1].is_trained)

        d3.work_dir = self.dirname
        d3.name = 'dummy-pipeline'
        d3.is_trained = True
        d3.dump_yaml()
        d3.dump()

        d4 = PipelineEncoder.load(d3.dump_full_path)
        self.assertTrue(d4.is_trained)
        self.assertTrue(d4.component[0].is_trained)
        self.assertTrue(d4.component[1].is_trained)

        d4 = PipelineEncoder.load_yaml(d3.yaml_full_path)
        self.assertTrue(d4.is_trained)
        self.assertTrue(d4.component[0].is_trained)
        self.assertTrue(d4.component[1].is_trained)

        self.assertEqual(d4.encode(4), 6)

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import List, Any, Union, Callable, Tuple

import numpy as np

from ..base import TrainableBase, CompositionalTrainableBase


class BaseIndexer(TrainableBase):

    def add(self, keys: Any, docs: Any, weights: List[float], *args, **kwargs):
        pass

    def query(self, keys: Any, *args, **kwargs) -> List[Any]:
        pass

    def normalize_score(self, *args, **kwargs):
        pass


class BaseVectorIndexer(BaseIndexer):

    def add(self, keys: List[Tuple[int, int]], vectors: np.ndarray, weights: List[float], *args, **kwargs):
        pass

    def query(self, keys: np.ndarray, top_k: int, *args, **kwargs) -> List[List[Tuple]]:
        pass

    def normalize_score(self, *args, **kwargs):
        pass


class BaseTextIndexer(BaseIndexer):

    def add(self, keys: List[int], docs: Any, weights: List[float], *args, **kwargs):
        pass

    def query(self, keys: List[int], *args, **kwargs) -> List[Any]:
        pass

    def normalize_score(self, *args, **kwargs):
        pass


class BaseKeyIndexer(BaseIndexer):

    def add(self, keys: List[Tuple[int, int]], weights: List[float], *args, **kwargs) -> int:
        pass

    def query(self, keys: List[int], *args, **kwargs) -> List[Tuple[int, int, float]]:
        pass

    def normalize_score(self, *args, **kwargs):
        pass


class JointIndexer(CompositionalTrainableBase):

    @property
    def component(self):
        return self._component

    @component.setter
    def component(self, comps: Callable[[], Union[list, dict]]):
        if not callable(comps):
            raise TypeError('component must be a callable function that returns '
                            'a List[BaseIndexer]')
        if not getattr(self, 'init_from_yaml', False):
            self._component = comps()
        else:
            self.logger.info('component is omitted from construction, '
                             'as it is initialized from yaml config')

        self._binary_indexer = None
        self._doc_indexer = None
        for c in self.component:
            if isinstance(c, BaseVectorIndexer):
                self._binary_indexer = c
            elif isinstance(c, BaseTextIndexer):
                self._doc_indexer = c
        if not self._binary_indexer or not self._doc_indexer:
            raise ValueError('"JointIndexer" requires a valid pair of "BaseBinaryIndexer" and "BaseTextIndexer"')

    def add(self, keys: Any, docs: Any, *args,
            **kwargs) -> None:
        if isinstance(docs, np.ndarray):
            self._binary_indexer.add(keys, docs, *args, **kwargs)
        elif isinstance(docs, list):
            self._doc_indexer.add(keys, docs, *args, **kwargs)
        else:
            raise TypeError('can not find an indexer for doc type: %s' % type(docs))

    def query(self,
              keys: Any,
              top_k: int,
              *args,
              **kwargs) -> List[List[Tuple]]:
        topk_results = self._binary_indexer.query(keys, top_k, *args, **kwargs)
        doc_caches = dict()
        topk_results_with_docs = []
        for topk in topk_results:
            topk_wd = []
            for doc_id, offset, weight, score in topk:
                doc = doc_caches.get(doc_id, self._doc_indexer.query([doc_id])[0])
                doc_caches[doc_id] = doc
                topk_wd.append((doc_id, offset, weight, score, doc.chunks[offset]))
            topk_results_with_docs.append(topk_wd)
        return topk_results_with_docs
